#!/usr/bin/env python3

# Command-line interface for the Network Control Framework of Zeek, using Broker.

import logging
import netcontrol
import sys
import _thread
from yaml import load
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader
import string
import re
import argparse
import sys

from subprocess import check_output
from subprocess import CalledProcessError
#from future.utils import viewitems

from enum import Enum, unique

def parseArgs():
    parser = argparse.ArgumentParser()
    parser.add_argument('--listen', default="127.0.0.1", help="Address to listen on for connections. Default: 127.0.0.1")
    parser.add_argument('--port', default=9977, help="Port to listen on for connections. Default: 9977")
    parser.add_argument('--topic', default="zeek/event/netcontrol-example", help="Topic to subscribe to. Default: zeek/event/netcontrol-example")
    parser.add_argument('--file', default="commands.yaml", help="File to read commands from. Default: commands.yaml")
    parser.add_argument('--debug', const=logging.DEBUG, default=logging.INFO, action='store_const', help="Enable debug output")

    args = parser.parse_args()
    return args

class Listen:
    def __init__(self, queue, host, port, commands, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.endpoint = netcontrol.Endpoint(queue, host, port)
        self.queuename = queue
        self.commands = commands
        self.use_threads = kwargs.get('use_threads', True)

    def listen_loop(self):
        self.logger.debug("Listen loop...")

        while 1==1:
            response = self.endpoint.getNextCommand()

            if response.type == netcontrol.ResponseType.AddRule:
                if self.use_threads:
                    _thread.start_new_thread(self._add_remove_rule, (response, ))
                else:
                    self._add_remove_rule(response)
            if response.type == netcontrol.ResponseType.RemoveRule:
                if self.use_threads:
                    _thread.start_new_thread(self._add_remove_rule, (response, ))
                else:
                    self._add_remove_rule(response)

    def _add_remove_rule(self, response):

        cmd = self.rule_to_cmd_dict(response.rule)

        if response.type == netcontrol.ResponseType.AddRule:
            type = 'add_rule'
        elif response.type == netcontrol.ResponseType.RemoveRule:
            type = 'remove_rule'
        else:
            self.logger.error("Internal error - incompabible rule type")
            type = 'unknown'

        if not ( type in self.commands):
            self.logger.error("No %s in commands", type)
            return

        commands = self.commands[type]

        output = ""

        self.logger.info("Received %s from Zeek: %s", type, cmd)

        for i in commands:
            currcmd = self.replace_command(i, cmd)
            output += "Command: "+currcmd+"\n"

            try:
                self.logger.info("Executing "+currcmd)
                cmdout = check_output(currcmd, shell=True)
                output += "Output: "+str(cmdout)+"\n"
                self.logger.debug("Command executed succefsully")
            except CalledProcessError as err:
                output = "Command "+currcmd+" failed with return code "+str(err.returncode)+" and output: "+str(err.output)
                self.logger.error(output)
                self.endpoint.sendRuleError(response, output)
                return
            except OSError as err:
                output = "Command "+currcmd+" failed with error code "+str(err.errno)+" ("+err.strerror+")"
                sel.logger.error(output)
                self.endpoint.sendRuleError(response, output)
                return

        if response.type == netcontrol.ResponseType.AddRule:
            self.logger.info("Sending rule_added to Zeek")
            self.endpoint.sendRuleAdded(response, output)
        else:
            self.logger.info("Sending rule_removed to Zeek")
            self.endpoint.sendRuleRemoved(response, output)

    def replace_single_command(self, argstr, cmds):
        reg = re.compile('\[(?P<type>.)(?P<target>.*?)(?:\:(?P<argument>.*?))?\]')
        #print argstr
        m = reg.search(argstr)

        if m == None:
            self.logger.error('%s could not be converted to rule', argstr)
            return ''

        type = m.group('type')
        target = m.group('target')
        arg = m.group('argument')

        if type == '?':
            if not ( target in cmds ):
                return ''
            elif arg == None:
                return cmds[target]

            # we have an argument *sigh*
            return re.sub(r'\.', cmds[target], arg)
        elif type == '!':
            if arg == None:
                self.logger.error("[!] needs argument for %s", argstr)
                return ''

            if not ( target in cmds ):
                return arg
            else:
                return ''
        else:
            self.logger.error("unknown command type %s in %s", type, argstr)
            return ''

    def replace_command(self, command, args):
        reg = re.compile('\[(?:\?|\!).*?\]')

        return reg.sub(lambda x: self.replace_single_command(x.group(), args), command)

    def rule_to_cmd_dict(self, rule):
        cmd = {}

        mapping = {
            'type': 'ty',
            'target': 'target',
            'expire': 'expire',
            'priority': 'priority',
            'id': 'id',
            'cid': 'cid',
            'entity.ip': 'address',
            'entity.mac': 'mac',
            'entity.conn.orig_h': 'conn.orig_h',
            'entity.conn.orig_p': 'conn.orig_p',
            'entity.conn.resp_h': 'conn.resp_h',
            'entity.conn.resp_p': 'conn.resp_p',
            'entity.flow.src_h': 'flow.src_h',
            'entity.flow.src_p': 'flow.src_p',
            'entity.flow.dst_h': 'flow.dst_h',
            'entity.flow.dst_p': 'flow.dst_p',
            'entity.flow.src_m': 'flow.src_m',
            'entity.flow.dst_m': 'flow.dst_m',
            'entity.mod.src_h': 'mod.src_h',
            'entity.mod.src_p': 'mod.src_p',
            'entity.mod.dst_h': 'mod.dst_h',
            'entity.mod.dst_p': 'mod.dst_p',
            'entity.mod.src_m': 'mod.src_m',
            'entity.mod.dst_m': 'mod.dst_m',
            'entity.mod.redirect_port': 'mod.port',
            'entity.i': 'mod.port',
        }

        for (k, v) in list(mapping.items()):
            path = k.split('.')
            e = rule
            for i in path:
                if e == None:
                    break
                elif i in e:
                    e = e[i]
                else:
                    e = None
                    break

            if e == None:
                continue

            if isinstance(e, tuple):
                cmd[v] = e[0]
                cmd[v+".proto"] = e[1]
            else:
                cmd[v] = e
                if isinstance(e, str):
                    spl = e.split("/")
                    if len(spl) > 1:
                        cmd[v+".ip"] = spl[0]
                        cmd[v+".net"] = spl[1]

        proto = mapping.get('entity.conn.orig_p.proto', mapping.get('entity.conn.dest_p.proto', mapping.get('entity.flow.src_p.proto', mapping.get('entity.flow.dst_p.proto', None))))
        if proto != None:
            entity['proto'] = proto

        return cmd

args = parseArgs()

stream = open(args.file, 'r')
config = load(stream, Loader=Loader)

logging.basicConfig(level=args.debug)

logging.info("Starting command-line client...")
zeekcon = Listen(args.topic, args.listen, int(args.port), config)
zeekcon.listen_loop()




from ZeekControl import version
from ZeekControl.ser import dumps
import os
import bottle
import json

app = bottle.Bottle(autojson=False)
app.install(bottle.JSONPlugin(json_dumps=lambda s: json.dumps(s, default=dumps)))

@app.route('/start')
def start():
    i = app.daemon.call("start")
    return {"id": i} 

@app.route('/stop')
def start():
    i = app.daemon.call("stop")
    return {"id": i} 

@app.route('/restart')
def restart():
    i = app.daemon.call("restart")
    return {"id": i} 


@app.route('/nodes')
def nodes():
    s = app.daemon.sync_call("nodes")
    print "\n\n\nReturning", s, "\n\n"
    return {"result": s}

@app.route('/exec/:cmd')
def start(cmd):
    i = app.daemon.call("execute", cmd)
    return {"id": i} 

@app.route('/result/:id')
def result(id):
    id = int(id)
    return {"result": app.daemon.getresult(id)}

@app.route('/log/:id')
@app.route('/log/:id/:since')
def result(id, since=0):
    id = int(id)
    since = int(since)
    return {"log": app.daemon.getlog(id, since) or []}

@app.route('/:cmd')
def cmd(cmd):
    i = app.daemon.call(cmd)
    return {"id": i} 

def run_app(client):
    app.daemon = client
    bottle.run(app, host='localhost', port=8082)

if __name__ == "__main__":
    main()




import unittest
import sys
import time

import broker

def create_stores():
    ep0 = broker.Endpoint()
    ep1 = broker.Endpoint()
    ep2 = broker.Endpoint()

    with ep0.make_subscriber("/test") as s0, \
         ep1.make_subscriber("/test") as s1, \
         ep1.make_status_subscriber() as es1, \
         ep2.make_subscriber("/test") as s2, \
         ep2.make_status_subscriber() as es2:

        p = ep0.listen("127.0.0.1", 0)
        ep1.peer("127.0.0.1", p)
        ep2.peer("127.0.0.1", p)

        # TODO: This doesn't work. Once it does, remove the event handshake.
        # es1.get()
        # es2.get()
        ep0.publish("/test", "go-ahead")
        s1.get()
        s2.get()
        ####

        m = ep0.attach_master("test", broker.Backend.Memory)
        c1 = ep1.attach_clone("test")
        c2 = ep2.attach_clone("test")

        return (ep0, ep1, ep2, m, c1, c2)

class TestStore(unittest.TestCase):
    def test_basic(self):
        # --master-start
        with broker.Endpoint() as ep1, \
             ep1.attach_master("test", broker.Backend.Memory) as m:

            m.put("key", "value")
            x = m.get("key")
            # x == "value"
            # --master-end
            self.assertEqual(x, "value")
            self.assertEqual(m.name(), "test")

    def test_from_master(self):
        (ep0, ep1, ep2, m, c1, c2) = create_stores()

        v1 = "A"
        v2 = {"A", "B", "C"}
        v3 = {1: "A", 2: "B", 3: "C"}
        v4 = ("A", "B", "C")

        m.put("a", v1)
        m.put("b", v2)
        m.put("c", v3)
        m.put("d", v4)
        self.assertEqual(c2.put_unique("e", "first"), True)
        self.assertEqual(c2.put_unique("e", "second"), False)
        self.assertEqual(c2.put_unique("e", "third"), False)
        time.sleep(.5)

        def checkAccessors(x):
            self.assertEqual(x.get("a"), v1)
            self.assertEqual(x.get("b"), v2)
            self.assertEqual(x.get("c"), v3)
            self.assertEqual(x.get("d"), v4)
            self.assertEqual(x.get("e"), "first")
            self.assertEqual(x.get("X"), None)
            self.assertEqual(x.exists("d"), True)
            self.assertEqual(x.exists("X"), False)
            self.assertEqual(x.get_index_from_value("b", "A"), True)
            self.assertEqual(x.get_index_from_value("b", "X"), False)
            self.assertEqual(x.get_index_from_value("c", 1), "A")
            self.assertEqual(x.get_index_from_value("c", 10), None)
            self.assertEqual(x.get_index_from_value("d", 1), "B")
            self.assertEqual(x.get_index_from_value("d", 10), None)
            self.assertEqual(x.keys(), {'a', 'b', 'c', 'd', 'e'})

        checkAccessors(m)
        checkAccessors(c1)
        checkAccessors(c2)

        v5 = 5
        m.put("e", v5)
        m.put("f", v5)
        m.put("g", v5, 0.1)
        m.put("h", v5, 2)
        m.put("str", "b")
        m.put("vec", (1, 2))
        m.put("set", set([1, 2]))
        m.put("table", {1: "A", "2": "C"})

        # --ops-start
        m.increment("e", 1)
        m.decrement("f", 1)
        m.append("str", "ar")
        m.insert_into("set", 3)
        m.remove_from("set", 1)
        m.insert_into("table", 3, "D")
        m.remove_from("table", 1)
        m.push("vec", 3)
        m.push("vec", 4)
        m.pop("vec")
        # --ops-end

        time.sleep(.5)

        def checkModifiers(x):
            self.assertEqual(x.get("e"), v5 + 1)
            self.assertEqual(x.get("f"), v5 - 1)
            self.assertEqual(x.get("g"), None) # Expired
            self.assertEqual(x.get("h"), v5) # Not Expired
            self.assertEqual(x.get("str"), "bar")
            self.assertEqual(x.get("set"), set([2, 3]))
            self.assertEqual(x.get("table"), {3: "D", "2": "C"})
            self.assertEqual(x.get("vec"), (1, 2, 3))

        checkModifiers(m)
        checkModifiers(c1)
        checkModifiers(c2)

        m.clear()
        time.sleep(.5)
        self.assertEqual(m.keys(), set())
        self.assertEqual(c1.keys(), set())
        self.assertEqual(c2.keys(), set())

        ep1.shutdown()
        ep2.shutdown()

    def test_from_clones(self):
        (ep0, ep1, ep2, m, c1, c2) = create_stores()

        v1 = "A"
        v2 = {"A", "B", "C"}
        v3 = {1: "A", 2: "B", 3: "C"}
        v4 = ("A", "B", "C")

        c1.put("a", v1)
        c1.put("b", v2)
        c2.put("c", v3)
        c2.put("d", v4)
        self.assertEqual(c2.put_unique("e", "first"), True)
        self.assertEqual(c2.put_unique("e", "second"), False)
        self.assertEqual(c2.put_unique("e", "third"), False)
        time.sleep(.5)

        def checkAccessors(x):
            self.assertEqual(x.get("a"), v1)
            self.assertEqual(x.get("b"), v2)
            self.assertEqual(x.get("c"), v3)
            self.assertEqual(x.get("d"), v4)
            self.assertEqual(x.get("e"), "first")
            self.assertEqual(x.get("X"), None)
            self.assertEqual(x.exists("d"), True)
            self.assertEqual(x.exists("X"), False)
            self.assertEqual(x.get_index_from_value("b", "A"), True)
            self.assertEqual(x.get_index_from_value("b", "X"), False)
            self.assertEqual(x.get_index_from_value("c", 1), "A")
            self.assertEqual(x.get_index_from_value("c", 10), None)
            self.assertEqual(x.get_index_from_value("d", 1), "B")
            self.assertEqual(x.get_index_from_value("d", 10), None)
            self.assertEqual(x.keys(), {'a', 'b', 'c', 'd', 'e'})

        checkAccessors(m)
        checkAccessors(c1)
        checkAccessors(c2)

        v5 = 5
        c1.put("e", v5)
        c2.put("f", v5)
        c1.put("g", v5, 0.1)
        c2.put("h", v5, 2)
        m.put("str", "b")
        m.put("vec", [1, 2])
        m.put("set", set([1, 2]))
        m.put("table", {1: "A", "2": "C"})

        time.sleep(.5)
        c2.increment("e", 1)
        c1.decrement("f", 1)
        c2.append("str", "ar")
        c1.insert_into("set", 3)
        c2.remove_from("set", 1)
        c1.insert_into("table", 3, "D")
        c2.remove_from("table", 1)
        c1.push("vec", 3)
        time.sleep(.5)
        c2.push("vec", 4)
        c2.pop("vec")
        time.sleep(.5)

        def checkModifiers(x):
            self.assertEqual(x.get("e"), v5 + 1)
            self.assertEqual(x.get("f"), v5 - 1)
            self.assertEqual(x.get("g"), None) # Expired
            self.assertEqual(x.get("h"), v5) # Not Expired
            self.assertEqual(x.get("str"), "bar")
            self.assertEqual(x.get("set"), set([2, 3]))
            self.assertEqual(x.get("table"), {3: "D", "2": "C"})
            self.assertEqual(x.get("vec"), (1, 2, 3))

        checkModifiers(m)
        checkModifiers(c1)
        checkModifiers(c2)

        m.clear()
        time.sleep(.5)
        self.assertEqual(m.keys(), set())
        self.assertEqual(c1.keys(), set())
        self.assertEqual(c2.keys(), set())

        ep1.shutdown()
        ep2.shutdown()


if __name__ == '__main__':
    unittest.main(verbosity=3)



#!/usr/bin/env python3

# Simple command line client using the provided API. Test, e.g., with
# the provided simple-test.zeek

import logging, netcontrol, pprint

logging.basicConfig(level=logging.DEBUG)

ep = netcontrol.Endpoint("zeek/event/netcontrol-example", "127.0.0.1", 9977);
pp = pprint.PrettyPrinter(indent=4)

while 1==1:
    response = ep.getNextCommand()

    if response.type == netcontrol.ResponseType.AddRule:
        ep.sendRuleAdded(response, "")
    elif response.type == netcontrol.ResponseType.RemoveRule:
        ep.sendRuleRemoved(response, "")
    else:
        continue

    pp.pprint(response.rule)



import json

from ZeekControl import node
from ZeekControl import cmdresult

class MyJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, node.Node):
            return obj.to_dict()
        if isinstance(obj, cmdresult.CmdResult):
            return obj.to_dict()
        return json.JSONEncoder.default(self, obj)

def dumps(obj):
    return json.dumps(obj, cls=MyJsonEncoder)



# sqlite-listen.py

import broker

with broker.Endpoint() as ep, \
     ep.make_subscriber('/test') as s, \
     ep.make_status_subscriber(True) as ss:

    ep.listen('127.0.0.1', 9999)

    m = ep.attach_master('mystore',
                         broker.Backend.SQLite, {'path': 'mystore.sqlite'})

    while True:
        print(ss.get())
        print(m.get('foo'))



import logging
import re
import ipaddress
import datetime
import broker
import broker.zeek
from select import select
from enum import Enum, unique

logger = logging.getLogger(__name__)

def convertRecord(name, m):
    if not isinstance(m, tuple):
        logger.error("Got non record element")

    rec = m

    elements = None
    if name == "rule":
        elements = ['ty', 'target', 'entity', 'expire', 'priority', 'location', 'out_port', 'mod', 'id', 'cid']
    elif name == "rule->entity":
        elements = ['ty', 'conn', 'flow', 'ip', 'mac']
    elif name == "rule->entity->conn":
        elements = ['orig_h', 'orig_p', 'resp_h', 'resp_p']
    elif name == "rule->entity->flow":
        elements = ['src_h', 'src_p', 'dst_h', 'dst_p', 'src_m', 'dst_m']
    elif name == "rule->mod":
        elements = ['src_h', 'src_p', 'dst_h', 'dst_p', 'src_m', 'dst_m', 'redirect_port']
    else:
        logger.error("Unknown record type %s", name)
        return

    dict = {}
    for i in range(0, len(elements)):
        if rec[i] is None:
            dict[elements[i]] = None
            continue
        elif isinstance(rec[i], tuple):
            dict[elements[i]] = convertRecord(name+"->"+elements[i], rec[i])
            continue

        dict[elements[i]] = convertElement(rec[i])

    return dict

def convertElement(el):
    if isinstance(el, broker.Count):
        return el.value

    if isinstance(el, ipaddress.IPv4Address):
        return str(el);

    if isinstance(el, ipaddress.IPv6Address):
        return str(el);

    if isinstance(el, ipaddress.IPv4Network):
        return str(el);

    if isinstance(el, ipaddress.IPv6Network):
        return str(el);

    if isinstance(el, broker.Port):
        p = str(el)
        ex = re.compile('([0-9]+)(.*)')
        res = ex.match(p)
        return (res.group(1), res.group(2))

    if isinstance(el, broker.Enum):
        tmp = el.name
        return re.sub(r'.*::', r'', tmp)

    if isinstance(el, tuple):
        return tuple(convertElement(ell) for ell in el);

    if isinstance(el, datetime.datetime):
        return el

    if isinstance(el, datetime.timedelta):
        return el

    if isinstance(el, int):
        return el

    if isinstance(el, str):
        return el

    logger.error("Unsupported type %s", type(el) )
    return el;

@unique
class ResponseType(Enum):
    ConnectionEstablished = 1
    Error = 2
    AddRule = 3
    RemoveRule = 4
    SelfEvent = 5

class NetControlResponse:
    def __init__(self):
        self.type = (ResponseType.Error)
        self.errormsg = ""
        self.rule = ""

    def __init__(self, rty, **kwargs):
        self.type = rty
        self.errormsg = kwargs.get('errormsg', '')
        self.pluginid = kwargs.get('pluginid', None)
        self.rule = kwargs.get('rule', None)
        self.rawrule = kwargs.get('rawrule', None)

class Endpoint:
    def __init__(self, queue, host, port):
        self.queuename = queue
        self.epl = broker.Endpoint()
        self.epl.listen(host, port)
        self.status_subscriber = self.epl.make_status_subscriber(True)
        self.subscriber = self.epl.make_subscriber(self.queuename)

        logger.debug("Set up listener for "+host+":"+str(port)+" ("+queue+")")
    def getNextCommand(self):
        while True:
            logger.debug("Waiting for broker message...")
            readable, writable, exceptional = select(
                    [self.status_subscriber.fd(), self.subscriber.fd()],
                    [], [])

            if ( self.status_subscriber.fd() in readable ):
                logger.debug("Handling broker status message...")
                msg = self.status_subscriber.get()

                if isinstance(msg, broker.Status):
                    if msg.code() == broker.SC.PeerAdded:
                        logger.info("Incoming connection established")
                        return NetControlResponse(ResponseType.ConnectionEstablished)

                    continue

            elif ( self.subscriber.fd() in readable ):
                logger.debug("Handling broker message...")
                msg = self.subscriber.get()
                return self.handleBrokerMessage(msg)

    def handleBrokerMessage(self, m):
        if type(m).__name__ != "tuple":
            logger.error("Unexpected type %s, expected tuple", type(m).__name__)
            return NetControlResponse(ResponseType.Error)

        if len(m) < 1:
            logger.error("Tuple without content?")
            return NetControlResponse(ResponseType.Error)

        (topic, event) = m
        ev = broker.zeek.Event(event)

        event_name = ev.name()
        logger.debug("Got event "+event_name)

        if event_name == "NetControl::broker_add_rule":
            return self._add_remove_rule(ev.args(), ResponseType.AddRule)
        elif event_name == "NetControl::broker_remove_rule":
            return self._add_remove_rule(ev.args(), ResponseType.RemoveRule)
        elif event_name == "NetControl::broker_rule_added":
            return NetControlResponse(ResponseType.SelfEvent)
        elif event_name == "NetControl::broker_rule_removed":
            return NetControlResponse(ResponseType.SelfEvent)
        elif event_name == "NetControl::broker_rule_error":
            return NetControlResponse(ResponseType.SelfEvent)
        elif event_name == "NetControl::broker_rule_timeout":
            return NetControlResponse(ResponseType.SelfEvent)
        else:
            logger.warning("Unknown event %s", event_name)
            return NetControlResponse(ResponseType.Error, errormsg="Unknown event"+event_name)

    def _add_remove_rule(self, m, rtype):
        if  ( (rtype == ResponseType.AddRule) and ( len(m) != 2 ) ) or ( (rtype == ResponseType.RemoveRule) and ( len(m) != 3 ) ):
            logger.error("wrong number of elements or type in tuple for add/remove_rule event")
            return NetControlResponse(ResponseType.Error, errormsg="wrong number of elements or type in tuple for add/remove_rule event")

        if ( not isinstance(m[0], broker.Count) or
             not isinstance(m[1], tuple) ):
            logger.error("wrong types of elements or type in tuple for add/remove_rule event")
            return NetControlResponse(ResponseType.Error, errormsg="wrong types of elements or type in tuple for add/remove_rule event")

        id = m[0].value
        rule = convertRecord("rule", m[1])

        return NetControlResponse(rtype, pluginid=id, rule=rule, rawrule=m[1])

    def sendRuleAdded(self, response, msg):
        self._rule_event("added", response, msg)

    def sendRuleRemoved(self, response, msg):
        self._rule_event("removed", response, msg)

    def sendRuleError(self, response, msg):
        self._rule_event("error", response, msg)

    def _rule_event(self, event, response, msg):
        args = [broker.Count(response.pluginid), response.rawrule, msg]
        ev = broker.zeek.Event("NetControl::broker_rule_"+event, *args)
        self.epl.publish(self.queuename, ev)



#!/usr/bin/env python

# This script demystifies C++ compiler output for CAF by
# replacing cryptic `typed_mpi<...>` templates with
# `replies_to<...>::with<...>` and `atom_constant<...>`
# with human-readable representation of the actual atom.

import sys

# decodes 6bit characters to ASCII
DECODING_TABLE = ' 0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz'

# CAF type strings
ATOM_CONSTANT_SUFFIX = "caf::atom_constant<"

# `pos` points to first character after '<':
#   template_name<...>
#                 ^
# and returns the position of the closing '>'
def end_of_template(x, pos):
  open_templates = 1
  while open_templates > 0:
    if line[pos] == '<':
      open_templates += 1
    elif line[pos] == '>':
      open_templates -= 1
    pos += 1
  # exclude final '>'
  return pos - 1

def next_element(x, pos, last):
  # scan for ',' that isn't inside <...>
  while pos < last and x[pos] != ',':
    if x[pos] == '<':
      pos = end_of_template(x, pos + 1)
    else:
      pos += 1
  return pos

def atom_read(x):
  result = ""
  read_chars = ((x & 0xF000000000000000) >> 60) == 0xF 
  mask = 0x0FC0000000000000
  bitshift = 54
  while bitshift >= 0:
    if read_chars:
      result += DECODING_TABLE[(x & mask) >> bitshift]
    elif ((x & mask) >> bitshift) == 0xF:
      read_chars = True
    bitshift -= 6
    mask = mask >> 6
  return result

def decompose_type_list(x, first, last):
  res = []
  i = first
  n = next_element(x, first, last)
  while i != last:
    res.append(x[i:n])
    # skip following ','
    i = min(n + 2, last)
    n = next_element(x, i, last)
  return res

def stringify(x):
  if x.startswith(ATOM_CONSTANT_SUFFIX):
    begin = len(ATOM_CONSTANT_SUFFIX)
    end = len(x) - 1
    res = "'"
    res += atom_read(int(x[begin:end]))
    res += "'"
    return res
  return x

def stringify_list(xs):
  res = ""
  for index in range(len(xs)):
    if index > 0:
      res += ", "
    res += stringify(xs[index].strip(' '))
  return res
  

def decompose_typed_actor(x, first, last):
  needle = "caf::detail::type_list<"
  # first type list -> input types
  j = x.find(needle, first) + len(needle)
  k = end_of_template(x, j)
  inputs = decompose_type_list(x, j, k)
  # second type list -> outputs
  j = x.find(needle, k) + len(needle)
  k = end_of_template(x, j)
  outputs = decompose_type_list(x, j, k)
  # replace all 'caf::atom_constant<...>' entries in inputs
  res = "replies_to<"
  res += stringify_list(inputs)
  res += ">::with<"
  res += stringify_list(outputs)
  res += ">"
  return res


for line in sys.stdin:
  # replace "std::__1" with "std::" (Clang libc++)
  line = line.replace("std::__1", "std::")
  needle = "caf::typed_mpi<"
  idx = line.find(needle)
  while idx != -1:
    # find end of typed_actor<...>
    first = idx + len(needle)
    last = end_of_template(line, first)
    updated = decompose_typed_actor(line, first, last)
    prefix = line[:idx]
    suffix = line[last:]
    line = prefix + updated + suffix
    idx = line.find(needle, idx + len(updated))
  sys.stdout.write(line.replace("caf::", ""))




# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

from datetime import date
import sys, os

# Add custom extensions directory to search path.
sys.path.insert(0, os.path.abspath('./extensions'))

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'numsec',
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Broker'
author = u'The Zeek Project'
copyright = str(date.today().year) + ', ' + author

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
with open('../VERSION', 'r') as f:
    version = f.readline().strip()

# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#
# today = ''
#
# Else, today_fmt is used as the format for a strftime call.
#
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', '.DS_Store']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

highlight_language = 'C++'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True
todo_emit_warnings = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:
    # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

html_theme_options = {
    'collapse_navigation': False,
    'display_version': True,
}

# The name for this set of Sphinx documents.
# "<project> v<release> documentation" by default.
#
html_title = u'Broker User Manual'

# A shorter title for the navigation bar.  Default is the same as html_title.
#
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# Custom sidebar templates, maps document names to template names.
#
# html_sidebars = {}
#html_sidebars = {
#   '**': ['custom-sidebar.html']
#}

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# Output file base name for HTML help builder.
htmlhelp_basename = 'broker-doc'

# -- Options for PDF output -----------------------------------------------

pdf_documents = [
    ('index', u'user-manual', u'Broker', u'The Zeek Project'),
]

pdf_stylesheets = ['sphinx', 'kerning', 'letter']



#!/usr/bin/env python3

# Ryu OpenFlow controller that connects to the Zeek OpenFlow
# framework using Broker.
#
# Start with ./ryu/bin/ryu-manager controller.py

import datetime
import ipaddress
import logging
import time
import re
import ryu.app.ofctl.api
from netaddr import IPNetwork

from ryu.base import app_manager
from ryu.controller import dpset
from ryu.controller import controller
from ryu.controller import ofp_event
from ryu.controller.handler import MAIN_DISPATCHER
from ryu.controller.handler import CONFIG_DISPATCHER
from ryu.controller.handler import set_ev_cls
from ryu.ofproto import ofproto_v1_0
from ryu.ofproto import ofproto_v1_2
from ryu.ofproto import ofproto_v1_3
from ryu.lib import ofctl_v1_0
from ryu.lib import ofctl_v1_2
from ryu.lib import ofctl_v1_3
from ryu.lib import hub

import broker
from select import select

supported_ofctl = {
    ofproto_v1_0.OFP_VERSION: ofctl_v1_0,
    ofproto_v1_2.OFP_VERSION: ofctl_v1_2,
    ofproto_v1_3.OFP_VERSION: ofctl_v1_3,
}

queuename = "bro/openflow"

# for monkey-patching.
# Barf.
def bro_send_msg(self, msg):
    assert isinstance(msg, self.ofproto_parser.MsgBase)

    if not hasattr(self, 'brosend'):
        self.brosend = 0
        self.bromessage = None

    # if we set before that we just want the message returned
    # without sending it on - return it to us so we can use
    # it further...
    if ( self.brosend == 1 ):
        self.brosend = 0
        self.bromessage = msg
        return msg

    self.send_msg_orig(msg)

# sorry about all that. This is just to get the API somewhere where we actually can work
# with it.
# :/
ryu.controller.controller.Datapath.send_msg_orig = ryu.controller.controller.Datapath.send_msg
ryu.controller.controller.Datapath.send_msg = bro_send_msg

class BroController(app_manager.RyuApp):

    OFP_VERSIONS = [ofproto_v1_0.OFP_VERSION,
                    ofproto_v1_2.OFP_VERSION,
                    ofproto_v1_3.OFP_VERSION]

    _CONTEXTS = {'dpset': dpset.DPSet}

    def __init__(self, *args, **kwargs):
        super(BroController, self).__init__(*args, **kwargs);

        self.dpset = kwargs['dpset']
        self.data = {}
        self.data['dpset'] = self.dpset
        # store DPID->name mapping. We create the mapping implicitely
        # for all flow_clear and flow_mod events that we get and use
        # it for the returning flow_removed events
        self.dpids = {}

        self.epl = broker.Endpoint()

    def start(self):
        self.epl.listen("127.0.0.1", 9999)
        self.status_subscriber = self.epl.make_status_subscriber(True)
        self.subscriber = self.epl.make_subscriber(queuename)

        self.threads.append(hub.spawn(self._broker_loop))
        self.logger.info("Started broker communication...")
        self.threads.append(hub.spawn(self._event_loop))

    @set_ev_cls(ofp_event.EventOFPSwitchFeatures, CONFIG_DISPATCHER)
    def _switch_features_handler(self, ev):
        dp = ev.msg.datapath
        # here we could alarm that we have seen a new switch

    def _broker_loop(self):
        self.logger.info("Broker loop...")

        while 1==1:
            self.logger.info("Waiting for broker message")
            readable, writable, exceptional = select(
                    [self.status_subscriber.fd(), self.subscriber.fd()],
                    [],[])

            if ( self.status_subscriber.fd() in readable ):
                self.logger.info("Got broker status message")
                msg = self.status_subscriber.get()
                self.handle_broker_message(msg)
            elif ( self.subscriber.fd() in readable ):
                self.logger.info("Got broker message")
                msg = self.subscriber.get()
                self.handle_broker_message(msg)

    def handle_broker_message(self, m):
        if isinstance(m, broker.Status):
            if m.code() == broker.SC.PeerAdded:
                self.logger.info("Incoming connection established.")
                return

            return

        if ( type(m).__name__ != "tuple" ):
            self.logger.error("Unexpected type %s, expected tuple", type(m).__name__)
            return

        if ( len(m) < 1 ):
            self.logger.error("Tuple without content?")
            return

        (topic, event) = m
        ev = broker.zeek.Event(event)
        event_name = ev.name()

        if ( event_name == "OpenFlow::broker_flow_clear" ):
            self.event_flow_clear(ev.args())
        elif ( event_name == "OpenFlow::broker_flow_mod" ):
            self.event_flow_mod(ev.args())
        elif event_name == "OpenFlow::flow_mod_success":
            pass
        elif event_name == "OpenFlow::flow_mod_failure":
            pass
        elif event_name == "OpenFlow::flow_removed":
            pass
        else:
            self.logger.error("Unknown event %s", event_name)
            return

    def event_flow_clear(self, m):
        if ( len(m) != 2 ) or ( not isinstance(m[0], str) ) or ( not isinstance(m[1], broker.Count) ):
            self.logger.error("wrong number of elements or type in tuple for event_flow_clear")
            return

        # since this is really only a  convenience function we should return it and just do the
        # flow-mod from Zeek ourselves
        name = m[0]

        dpid = m[1].value
        self.logger.info("flow_clear for %s %d", name, dpid)

        dp = ryu.app.ofctl.api.get_datapath(self, int(dpid))

        if dp is None:
            self.logger.error("dpid %d not found for clear", dpid)
            return

        self.dpids[dp.id] = name

        flow = {'table_id': dp.ofproto.OFPTT_ALL}
        _ofp_version = dp.ofproto.OFP_VERSION
        _ofctl = supported_ofctl.get(_ofp_version, None)
        if _ofctl is None:
            self.logger.error("unsupported openflow protocol")
            return

        dp.brosend = 1 # give it to us
        _ofctl.mod_flow_entry(dp, flow, dp.ofproto.OFPFC_DELETE)
        msg = dp.bromessage

        ryu.app.ofctl.api.send_msg(self, msg)

    def send_error(self, name, match, flow_mod, msg):
        ev = broker.zeek.Event("OpenFlow::flow_mod_failure", name, match, flow_mod, msg)
        self.epl.publish(queuename, ev)

    def send_success(self, name, match, flow_mod, msg):
        ev = broker.zeek.Event("OpenFlow::flow_mod_success", name, match, flow_mod, msg)
        self.epl.publish(queuename, ev)

    def event_flow_mod(self, m):
        if ( len(m) != 4 ) or ( not isinstance(m[0], str) ) or ( not isinstance(m[1], broker.Count) ) or ( not isinstance(m[2], tuple) ) or ( not isinstance(m[3], tuple) ):
            self.logger.error("wrong number of elements or type in tuple for event_flow_mod")
            return

        name = m[0]

        dpid = m[1].value
        match = self.parse_ofp_match(m[2])
        flow_mod = self.parse_ofp_flow_mod(m[3])

        dp = ryu.app.ofctl.api.get_datapath(self, int(dpid))

        if dp is None:
            self.logger.error("name %s dpid %d not found for flow_mod", name, dpid)
            self.send_error(name, m[2], m[3], "dpid not found")
            return

        self.dpids[dp.id] = name

        if dp.ofproto.OFP_VERSION != ofproto_v1_0.OFP_VERSION:
            if 'nw_dst' in match:
                if ":" in match['nw_dst']:
                    match['ipv6_dst'] = match['nw_dst']
                else:
                    match['ipv4_dst'] = match['nw_dst']
                del match['nw_dst']

            if 'nw_src' in match:
                if ":" in match['nw_src']:
                    match['ipv6_src'] = match['nw_src']
                else:
                    match['ipv4_src'] = match['nw_src']
                del match['nw_src']

            if 'tp_src' in match:
                proto = match.get('nw_proto', None);
                if proto == None:
                    self.logger.error("Cannot determine proto for flow mod")
                    return

                if proto == 0x06:
                    match['tcp_src'] = match['tp_src']
                    del match['tp_src']
                elif proto == 0x11:
                    match['udp_src'] = match['tp_src']
                    del match['tp_src']
                elif proto == 0x01:
                    match['icmpv4_type'] = match['tp_src']
                    del match['tp_src']


            if 'tp_dst' in match:
                proto = match.get('nw_proto', None);
                if proto == None:
                    self.logger.error("Cannot determine proto for flow mod")
                    return

                if proto == 0x06:
                    match['tcp_dst'] = match['tp_dst']
                    del match['tp_dst']
                elif proto == 0x11:
                    match['udp_dst'] = match['tp_dst']
                    del match['tp_dst']
                elif proto == 0x01:
                    match['icmpv4_code'] = match['tp_dst']
                    del match['tp_dst']

        self.logger.info("flow_mod for %d", dpid)
        #print match
        #print flow_mod


        cmdstr = flow_mod['command']
        cmd = self.string_to_command(dp, cmdstr)

        flow_mod['match'] = match
        #flow_mod['flags'] = 1 # remove, we actually want overlapping entries sometimes.
        actions = []

        if dp.ofproto.OFP_VERSION == ofproto_v1_0.OFP_VERSION:
            for k, v in flow_mod['actions'].iteritems():
                if k == 'vlan_vid':
                    actions.append(dp.ofproto_parser.OFPActionVlanVid(v))
                elif k == 'vlan_pcp':
                    actions.append(dp.ofproto_parser.OFPActionVlanPcp(k))
                elif k == 'vlan_strip' and v == True:
                    actions.append(dp.ofproto_parser.OFPActionStripVlan())
                elif k == 'dl_src':
                    dl_src = haddr_to_bin(v)
                    actions.append(dp.ofproto_parser.OFPActionSetDlSrc(dl_src))
                elif k == 'dl_dst':
                    dl_dst = haddr_to_bin(v)
                    actions.append(dp.ofproto_parser.OFPActionSetDlDst(dl_dst))
                elif k == 'nw_tos':
                    actions.append(dp.ofproto_parser.OFPActionSetNwTos(v))
                elif k == 'nw_src':
                    actions.append(dp.ofproto_parser.OFPActionSetNwSrc(ipv4_to_int(v)))
                elif k == 'nw_dst':
                    actions.append(dp.ofproto_parser.OFPActionSetNwDst(ipv4_to_int(v)))
                elif k == 'tp_src':
                    actions.append(dp.ofproto_parser.OFPActionSetTpSrc(v))
                elif k == 'tp_dst':
                    actions.append(dp.ofproto_parser.OFPActionSetTpDst(v))
        else:

            if 'nw_dst' in flow_mod['actions']:
                if ":" in flow_mod['actions']['nw_dst']:
                    flow_mod['actions']['ipv6_dst'] = flow_mod['actions']['nw_dst']
                else:
                    flow_mod['actions']['ipv4_dst'] = flow_mod['actions']['nw_dst']
                del flow_mod['actions']['nw_dst']

            if 'nw_src' in flow_mod['actions']:
                if ":" in flow_mod['actions']['nw_src']:
                    flow_mod['actions']['ipv6_src'] = flow_mod['actions']['nw_src']
                else:
                    flow_mod['actions']['ipv4_src'] = flow_mod['actions']['nw_src']
                del flow_mod['actions']['nw_src']

            if 'tp_src' in flow_mod['actions']:
                proto = match.get('nw_proto', None);
                if proto == None:
                    self.logger.error("Cannot determine proto for flow mod")
                    return

                if proto == 0x06:
                    flow_mod['actions']['tcp_src'] = flow_mod['actions']['tp_src']
                elif proto == 0x11:
                    flow_mod['actions']['udp_src'] = flow_mod['actions']['tp_src']
                elif proto == 0x01:
                    flow_mod['actions']['icmpv4_type'] = flow_mod['actions']['tp_src']

            if 'tp_dst' in flow_mod['actions']:
                proto = match.get('nw_proto', None);
                if proto == None:
                    self.logger.error("Cannot determine proto for flow mod")
                    return

                if proto == 0x06:
                    flow_mod['actions']['tcp_dst'] = flow_mod['actions']['tp_dst']
                elif proto == 0x11:
                    flow_mod['actions']['udp_dst'] = flow_mod['actions']['tp_dst']
                elif proto == 0x01:
                    flow_mod['actions']['icmpv4_code'] = flow_mod['actions']['tp_dst']

            for k, v in flow_mod['actions'].iteritems():
                if k == 'vlan_strip' and v == True:
                    actions.append(dp.ofproto_parser.OFPActionStripVlan())
                elif ( k == 'vlan_vid' ) or ( k == 'vlan_pcp' ) or ( k == 'nw_tos' ) or ( k == 'ipv4_src' ) or ( k == 'ipv4_dst' ) or ( k == 'tcp_src' ) or ( k == 'tcp_dst' ) or ( k == 'udp_src' ) or ( k == 'udp_dst' ) or ( k == 'icmpv4_code' ) or ( k == 'icmpv4_type' ) or ( k == 'ipv6_src' ) or ( k == 'ipv6_dst' ):
                    pass
                    actions.append(dp.ofproto_parser.OFPActionSetField(**{k: v}))
                elif ( k == 'dl_src' ) or ( k == 'dl_dst' ):
                    #dl = haddr_to_bin(v)
                    actions.append(dp.ofproto_parser.OFPActionSetField(**{k: v}))

        # do out-ports separately because it has to be last...
        if 'out_ports' in flow_mod['actions']:
            for i in flow_mod['actions']['out_ports']:
                max_len = 0xffe5
                if dp.ofproto.OFP_VERSION != ofproto_v1_0.OFP_VERSION:
                    max_len = dp.ofproto.OFPCML_MAX

                if dp.ofproto.OFP_VERSION == ofproto_v1_0.OFP_VERSION:
                    if i == 0xfffffff8:
                        i = dp.ofproto.OFPP_IN_PORT
                    elif i == 0xfffffff9:
                        i = dp.ofproto.OFPP_TABLE
                    elif i == 0xfffffffa:
                        i = dp.ofproto.OFPP_NORMAL
                    elif i == 0xfffffffb:
                        i = dp.ofproto.OFPP_FLOOD
                    elif i == 0xfffffffc:
                        i = dp.ofproto.OFPP_ALL
                    elif i == 0xfffffffd:
                        i = dp.ofproto.OFPP_CONTROLLER
                    elif i == 0xfffffffe:
                        i = dp.ofproto.OFPP_LOCAL
                    elif i == 0xffffffff:
                        i = dp.ofproto.OFPP_ANY

                actions.append(dp.ofproto_parser.OFPActionOutput(i, max_len))

        del flow_mod['actions']

        if cmd is None:
            self.logger.error("command %s could not be parsed", cmdstr)
            self.send_error(m[2], m[3], "cmd not recognized")
            return

        _ofp_version = dp.ofproto.OFP_VERSION
        _ofctl = supported_ofctl.get(_ofp_version, None)
        if _ofctl is None:
            self.logger.error("unsupported openflow protocol")
            self.send_error(m[2], m[3], "unsupported openflow protocol")
            return

        dp.brosend = 1 # give it to us...
        _ofctl.mod_flow_entry(dp, flow_mod, cmd)
        msg = dp.bromessage

        # naming and calling changed in the api for of1.0 vs 1.3

        insts = []
        if dp.ofproto.OFP_VERSION == ofproto_v1_0.OFP_VERSION:
            msg.actions = actions
        else:
            insts.append(dp.ofproto_parser.OFPInstructionActions(dp.ofproto.OFPIT_APPLY_ACTIONS, actions))
            msg.instructions = insts

        #print "Sending to switch:"
        #print msg

        try:
            ryu.app.ofctl.api.send_msg(self, msg)
            self.send_success(name, m[2], m[3], "")
        except ryu.app.ofctl.exception.OFError as err:
            self.logger.error("flow_mod execution error %s", err)
            self.send_error(name, m[2], m[3], str(err))

    def parse_ofp_match(self, m):
        match = ['in_port', 'dl_src', 'dl_dst', 'dl_vlan', 'dl_vlan_pcp', 'dl_type', 'nw_tos', 'nw_proto', 'nw_src', 'nw_dst', 'tp_src', 'tp_dst']
        return self.record_to_record(match, m)


    def parse_ofp_flow_mod(self, m):
        match = ['cookie', 'table_id', 'command', 'idle_timeout', 'hard_timeout', 'priority', 'out_port', 'out_group', 'flags']

        rec = self.record_to_record(match, m)

        # ok, now we have to get the actions, which are after flags. This is kind of cheating, but... whatever :)
        match_actions = ['out_ports', 'vlan_vid', 'vlan_pcp', 'vlan_strip', 'dl_src', 'dl_dst', 'nw_tos', 'nw_src', 'nw_dst', 'tp_src', 'tp_dst']

        rm = m
        rl = rm[9]
        recaction = self.record_to_record(match_actions, rl)
        rec['actions'] = recaction

        return rec

    @set_ev_cls(ofp_event.EventOFPFlowRemoved, MAIN_DISPATCHER)
    def _flow_removed_handler(self, ev):
        msg = ev.msg
        dp = msg.datapath
        ofp = dp.ofproto
        match = msg.match

        if dp.id not in self.dpids:
            self.logger.error("Flow remove for unknown DPID %d", dp.id)
            return

        #print "Flow removed"

        match_vec = vector_of_field([])
        match_vec = self.vec_add_field(match_vec, match, 'in_port');
        match_vec = self.vec_add_field(match_vec, match, 'eth_src');
        match_vec = self.vec_add_field(match_vec, match, 'eth_dst');
        match_vec = self.vec_add_field(match_vec, match, 'vlan_vid');
        match_vec = self.vec_add_field(match_vec, match, 'vlan_pcp');
        match_vec = self.vec_add_field(match_vec, match, 'eth_type');
        match_vec = self.vec_add_field(match_vec, match, 'ip_dscp');
        match_vec = self.vec_add_field(match_vec, match, 'ip_proto');

        src = match.get('ipv4_src', match.get('ipv6_src', match.get('nw_src', None)))
        dst = match.get('ipv4_dst', match.get('ipv6_dst', match.get('nw_dst', None)))

        if src != None:
            sn = None
            if ( not isinstance(src, tuple) ) and( ":" in src ):
                sn = subnet(address.from_string(src), 128)
            elif not isinstance(src, tuple):
                sn = subnet(address.from_string(src), 32)
            else:
                adr = address.from_string(src[0])
                ip = IPNetwork(src[0]+"/"+src[1])
                sn = subnet(adr, ip.prefixlen)
            match_vec.push_back(field(data(sn)))
        else:
            match_vec.push_back(field())

        if dst != None:
            sn = None
            if ( not isinstance(dst, tuple) ) and( ":" in dst ):
                sn = subnet(address.from_string(dst), 128)
            elif not isinstance(dst, tuple):
                sn = subnet(address.from_string(dst), 32)
            else:
                adr = address.from_string(dst[0])
                ip = IPNetwork(dst[0]+"/"+dst[1])
                sn = subnet(adr, ip.prefixlen)
            match_vec.push_back(field(data(sn)))
        else:
            match_vec.push_back(field())

        srcp = match.get('tcp_src', match.get('udp_src', match.get('icmpv4_type', match.get('tp_src', None))))
        dstp = match.get('tcp_dst', match.get('udp_dst', match.get('icmpv4_type', match.get('tp_dst', None))))

        if srcp != None:
            match_vec.push_back(field(data(srcp)))
        else:
            match_vec.push_back(field())

        if dstp != None:
            match_vec.push_back(field(data(dstp)))
        else:
            match_vec.push_back(field())

        args = [self.dpids[dp.id],
                match_vec,
                broker.Count(msg.cookie),
                broker.Count(msg.priority),
                broker.Count(msg.reason),
                broker.Count(msg.duration_sec),
                broker.Count(msg.idle_timeout),
                broker.Count(msg.packet_count),
                broker.Count(msg.byte_count)]
        ev = broker.zeek.Event("OpenFlow::flow_removed", args)
        self.epl.publish(queuename, ev)


    def vec_add_field(self, match_vec, match , el):
        if el in match:
            match_vec.push_back(field(data(match[el])))
        else:
            match_vec.push_back(field())

        return match_vec

    def string_to_command(self, dp, cmdstr):
        if ( cmdstr == "OFPFC_ADD" ):
            return dp.ofproto.OFPFC_ADD
        elif ( cmdstr == "OFPFC_MODIFY" ):
            return dp.ofproto.OFPFC_MODIFY
        elif ( cmdstr == "OFPFC_MODIFY_STRICT" ):
            return dp.ofproto.OFPFC_MODIFY_STRICT
        elif ( cmdstr == "OFPFC_DELETE" ):
            return dp.ofproto.OFPFC_DELETE
        elif ( cmdstr == "OFPFC_DELETE_STRICT" ):
            return dp.ofproto.OFPFC_DELETE_STRICT
        else:
            return None

    def record_to_record(self, match, m):
        #if len(match) != len(m):
        #    self.logger.error("wrong number of elements in parse_ofp_match")
        #    return

        if not isinstance(m, tuple):
            self.logger.error("Got non record element")

        rec = m

        dict = {}
        for i in range(0, len(match)):
            if rec[i] is None:
                #dict[match[i]] = None # most of the functions expect this to be undefined, not none. We oblige.
                continue

            dict[match[i]] = self.convert_element(rec[i])

        return dict

    def convert_element(self, el):
        if isinstance(el, broker.Count):
            return el.value

        if isinstance(el, ipaddress.IPv4Address):
            return str(el);

        if isinstance(el, ipaddress.IPv6Address):
            return str(el);

        if isinstance(el, ipaddress.IPv4Network):
            return str(el);

        if isinstance(el, ipaddress.IPv6Network):
            return str(el);

        if isinstance(el, broker.Port):
            p = str(el)
            ex = re.compile('([0-9]+)(.*)')
            res = ex.match(p)
            return (res.group(1), res.group(2))

        if isinstance(el, broker.Enum):
            tmp = el.name
            return re.sub(r'.*::', r'', tmp)

        if isinstance(el, tuple):
            return tuple(self.convert_element(ell) for ell in el);

        if isinstance(el, datetime.datetime):
            return el

        if isinstance(el, datetime.timedelta):
            return el

        if isinstance(el, int):
            return el

        if isinstance(el, str):
            return el

        logger.error("Unsupported type %s", type(el) )
        return el;



# This plugin sets necessary environment variables to run Zeek with
# PF_RING load balancing.

import ZeekControl.plugin
import ZeekControl.config

class LBPFRing(ZeekControl.plugin.Plugin):
    def __init__(self):
        super(LBPFRing, self).__init__(apiversion=1)

    def name(self):
        return "lb_pf_ring"

    def pluginVersion(self):
        return 1

    def init(self):
        cluster_id = ZeekControl.config.Config.pfringclusterid
        if cluster_id == 0:
            return False

        pfringtype = ZeekControl.config.Config.pfringclustertype
        if pfringtype not in ("2-tuple", "4-tuple", "5-tuple", "tcp-5-tuple",
            "6-tuple", "round-robin"):
            self.error("Invalid configuration: PFRINGClusterType=%s" % pfringtype)
            return False

        # If the cluster type is not round-robin, then choose the corresponding
        # environment variable.
        pftype = ""
        if pfringtype != "round-robin":
            pftype = "PCAP_PF_RING_USE_CLUSTER_PER_FLOW"
            if pfringtype != "6-tuple":
                pftype += "_" + pfringtype.upper().replace("-", "_")

        useplugin = False
        first_app_instance = ZeekControl.config.Config.pfringfirstappinstance
        app_instance = first_app_instance

        dd = {}
        for nn in self.nodes():
            if nn.type != "worker" or nn.lb_method != "pf_ring":
                continue

            useplugin = True

            if nn.host in dd:
                if nn.interface not in dd[nn.host]:
                    app_instance = first_app_instance
                    dd[nn.host][nn.interface] = cluster_id + len(dd[nn.host])
            else:
                app_instance = first_app_instance
                dd[nn.host] = {nn.interface: cluster_id}

            # Apply environment variables, but do not override values from
            # the node.cfg or zeekctl.cfg files.
            if pftype:
                nn.env_vars.setdefault(pftype, "1")

            if nn.interface.startswith("zc:"):
                # For the case where a user is doing RSS with ZC or
                # load-balancing with zbalance_ipc (through libpcap over
                # pf_ring)
                nn.env_vars.setdefault("PCAP_PF_RING_ZC_RSS", "1")
                nn.interface = "%s@%d" % (nn.interface, app_instance)

            elif nn.interface.startswith("pf_ring::zc:"):
                # For the case where a user is doing RSS with ZC or
                # load-balancing with zbalance_ipc (through the zeek::pf_ring
                # plugin)
                nn.env_vars.setdefault("PCAP_PF_RING_ZC_RSS", "1")
                nn.interface = "%s@%d" % (nn.interface, app_instance)

            elif nn.interface.startswith("dnacl"):
                # For the case where a user is running pfdnacluster_master (deprecated)
                nn.interface = "%s@%d" % (nn.interface, app_instance)

            elif nn.interface.startswith("dna"):
                # For the case where a user is doing symmetric RSS with DNA (deprecated)
                nn.env_vars.setdefault("PCAP_PF_RING_DNA_RSS", "1")
                nn.interface = "%s@%d" % (nn.interface, app_instance)

            else:
                nn.env_vars.setdefault("PCAP_PF_RING_CLUSTER_ID", dd[nn.host][nn.interface])

            app_instance += 1
            nn.env_vars.setdefault("PCAP_PF_RING_APPNAME", "zeek-%s" % nn.interface)

        return useplugin




# ping.py

import sys
import broker

# Setup endpoint and connect to Zeek.
with broker.Endpoint() as ep, \
     ep.make_subscriber("/topic/test") as sub, \
     ep.make_status_subscriber(True) as ss:

    ep.peer("127.0.0.1", 9999)

    # Wait until connection is established.
    st = ss.get()

    if not (type(st) == broker.Status and st.code() == broker.SC.PeerAdded):
        print("could not connect")
        sys.exit(0)

    for n in range(5):
        # Send event "ping(n)".
        ping = broker.zeek.Event("ping", n);
        ep.publish("/topic/test", ping);

        # Wait for "pong" reply event.
        (t, d) = sub.get()
        pong = broker.zeek.Event(d)
        print("received {}{}".format(pong.name(), pong.args()))



# Build the DebugCmdConstants.h and DebugCmdInfoConstants.cc files from the
# DebugCmdInfoConstants.in file.
#
# We do this via a script rather than maintaining them directly because
# the struct is a little complicated, so has to be initialized from code,
# plus we want to make adding new constants somewhat less painful.
#
# The input filename should be supplied as an argument.
#
# DebugCmds are printed to DebugCmdConstants.h
# DebugCmdInfos are printed to DebugCmdInfoConstants.cc
#
# The input format is:
#
#	cmd: [DebugCmd]
#	names: [space delimited names of cmd]
#	resume: ['true' or 'false': should execution resume after this command?]
#	help: [some help text]
#
# Blank lines are skipped.
# Comments should start with // and should be on a line by themselves.

import sys

inputfile = sys.argv[1]

init_tmpl = '''
\t{
\t\tDebugCmdInfo* info;
\t\t%(name_init)s
\t\tinfo = new DebugCmdInfo(%(cmd)s, names, %(num_names)s, %(resume)s, "%(help)s",
\t\t                                      %(repeatable)s);
\t\tg_DebugCmdInfos.push_back(info);
\t}
'''

enum_str = '''
//
// This file was automatically generated from %s
// DO NOT EDIT.
//
enum DebugCmd {
''' % inputfile

init_str = '''
//
// This file was automatically generated from %s
// DO NOT EDIT.
//

#include "zeek/util.h"
namespace zeek::detail {\n
void init_global_dbg_constants () {
''' % inputfile


def outputrecord():
    global init_str, enum_str

    if dbginfo["names"]:
        dbginfo["name_init"] = "const char * const names[] = {\n"\
                               "\t\t\t%s\n"\
                               "\t\t};\n" \
                               % ",\n\t\t\t".join(dbginfo["names"])
    else:
        dbginfo["name_init"] = "const char * const names[] = { };\n"

    dbginfo["num_names"] = len(dbginfo["names"])

    # substitute into template
    init_str += init_tmpl % dbginfo

    enum_str += "\t%s,\n" % dbginfo["cmd"]


def initdbginfo():
    return {
        "cmd": "",
        "name_init": "",
        "num_names": 0,
        "names": [],
        "resume": "false",
        "help": "",
        "repeatable": "false"
    }


dbginfo = initdbginfo()

inputf = open(inputfile, "r")
for line in inputf:
    line = line.strip()
    if not line or line.startswith("//"):  # skip empty lines and comments
        continue

    fields = line.split(":", 1)
    if len(fields) != 2:
        raise RuntimeError("Error in debug constant file on line: %s" % line)

    f1, f2 = fields
    f2 = f2.strip()

    if f1 == "cmd":
        if dbginfo[f1]:  # output the previous record
            outputrecord()
            dbginfo = initdbginfo()

        dbginfo[f1] = f2
    elif f1 == "names":
        # put quotes around the strings
        dbginfo[f1] = ['"%s"' % n for n in f2.split()]
    elif f1 == "help":
        dbginfo[f1] = f2.replace('"', '\\"')  # escape quotation marks
    elif f1 in ("resume", "repeatable"):
        dbginfo[f1] = f2
    else:
        raise RuntimeError("Unknown command: %s" % line)

# output the last record
outputrecord()

init_str += "\t\n}\n\n} // namespace zeek::detail\n"
enum_str += "\tdcLast\n};\n"

debugcmds = open("DebugCmdConstants.h", "w")
debugcmds.write(enum_str)
debugcmds.close()

debugcmdinfos = open("DebugCmdInfoConstants.cc", "w")
debugcmdinfos.write(init_str)
debugcmdinfos.close()



import unittest
import multiprocessing
import sys
import time
import os.path

import broker

def data_path(file):
    base = os.path.realpath(__file__)
    return os.path.join(os.path.join(os.path.dirname(base), "certs"), file)

class TestSSL(unittest.TestCase):
    def check_ping(self, ep1, s1, ep2, s2):
        ep2.publish("/test", ["ping"])
        (t, d) = s1.get()
        self.assertEqual(t, "/test")
        self.assertEqual(d[0], "ping")

        ep1.publish(t, ["pong"])
        (t, d) = s2.get()
        self.assertEqual(t, "/test")
        self.assertEqual(d[0], "pong")

    def test_ssl_auth_success_ca(self):
        cfg = broker.Configuration(broker.BrokerOptions())
        cfg.openssl_certificate = data_path("cert.1.pem")
        cfg.openssl_key = data_path("key.1.pem")
        cfg.openssl_cafile = data_path("ca.pem")

        with broker.Endpoint(cfg) as ep1, \
             broker.Endpoint(cfg) as ep2, \
             ep1.make_subscriber("/test") as s1, \
             ep2.make_subscriber("/test") as s2:

            port = ep1.listen("127.0.0.1", 0)
            r = ep2.peer("127.0.0.1", port, 0)
            self.assertEqual(r, True)

            self.check_ping(ep1, s1, ep2, s2)

    def test_ssl_auth_success_ca_pw(self):
        cfg = broker.Configuration(broker.BrokerOptions())
        cfg.openssl_certificate = data_path("cert.1.pem")
        cfg.openssl_key = data_path("key.1.enc.pem")
        cfg.openssl_cafile = data_path("ca.pem")
        cfg.openssl_passphrase = "12345"

        with broker.Endpoint(cfg) as ep1, \
             broker.Endpoint(cfg) as ep2, \
             ep1.make_subscriber("/test") as s1, \
             ep2.make_subscriber("/test") as s2:

            port = ep1.listen("127.0.0.1", 0)
            r = ep2.peer("127.0.0.1", port, 0)
            self.assertEqual(r, True)

            self.check_ping(ep1, s1, ep2, s2)

    def test_ssl_auth_success_self_signed(self):
        cfg = broker.Configuration(broker.BrokerOptions())
        cfg.openssl_certificate = data_path("cert.self-signed.pem")
        cfg.openssl_key = data_path("key.self-signed.pem")
        cfg.openssl_cafile = data_path("cert.self-signed.pem")

        with broker.Endpoint(cfg) as ep1, \
             broker.Endpoint(cfg) as ep2, \
             ep1.make_subscriber("/test") as s1, \
             ep2.make_subscriber("/test") as s2:

            port = ep1.listen("127.0.0.1", 0)
            r = ep2.peer("127.0.0.1", port, 0)
            self.assertEqual(r, True)

            self.check_ping(ep1, s1, ep2, s2)

    def test_ssl_auth_failure_self_signed(self):
        cfg1 = broker.Configuration(broker.BrokerOptions())
        cfg1.openssl_certificate = data_path("cert.1.pem")
        cfg1.openssl_key = data_path("key.1.pem")
        cfg1.openssl_cafile = data_path("ca.pem")

        cfg2 = broker.Configuration(broker.BrokerOptions())
        cfg2.openssl_certificate = data_path("cert.self-signed.pem")
        cfg2.openssl_key = data_path("key.self-signed.pem")
        cfg2.openssl_cafile = data_path("cert.self-signed.pem")

        with broker.Endpoint(cfg1) as ep1, \
             broker.Endpoint(cfg2) as ep2:

            port = ep1.listen("127.0.0.1", 0)
            r = ep2.peer("127.0.0.1", port, 0)
            self.assertEqual(r, False)

        with broker.Endpoint(cfg2) as ep1, \
             broker.Endpoint(cfg1) as ep2:

            port = ep1.listen("127.0.0.1", 0)
            r = ep2.peer("127.0.0.1", port, 0)
            self.assertEqual(r, False)


    def test_ssl_auth_failure_no_auth(self):
        cfg1 = broker.Configuration(broker.BrokerOptions())
        cfg1.openssl_certificate = data_path("cert.1.pem")
        cfg1.openssl_key = data_path("key.1.pem")
        cfg1.openssl_cafile = data_path("ca.pem")

        cfg2 = broker.Configuration(broker.BrokerOptions())

        with broker.Endpoint(cfg1) as ep1, \
             broker.Endpoint(cfg2) as ep2:

            port = ep1.listen("127.0.0.1", 0)
            r = ep2.peer("127.0.0.1", port, 0)
            self.assertEqual(r, False)

        with broker.Endpoint(cfg2) as ep1, \
             broker.Endpoint(cfg1) as ep2:

            port = ep1.listen("127.0.0.1", 0)
            r = ep2.peer("127.0.0.1", port, 0)
            self.assertEqual(r, False)

    def test_ssl_auth_failure_no_ssl(self):
        cfg1 = broker.Configuration(broker.BrokerOptions())
        cfg1.openssl_certificate = data_path("cert.1.pem")
        cfg1.openssl_key = data_path("key.1.pem")
        cfg1.openssl_cafile = data_path("ca.pem")

        cfg2 = broker.Configuration(broker.BrokerOptions())

        with broker.Endpoint(cfg1) as ep1, \
             broker.Endpoint(cfg2) as ep2:

            port = ep1.listen("127.0.0.1", 0)
            r = ep2.peer("127.0.0.1", port, 0)
            self.assertEqual(r, False)

        with broker.Endpoint(cfg2) as ep1, \
             broker.Endpoint(cfg1) as ep2:

            port = ep1.listen("127.0.0.1", 0)
            r = ep2.peer("127.0.0.1", port, 0)
            self.assertEqual(r, False)

    def XXXtest_ssl_auth_failure_ca_pw(self):
        cfg = broker.Configuration(broker.BrokerOptions())
        cfg.openssl_certificate = data_path("cert.1.pem")
        cfg.openssl_key = data_path("key.1.enc.pem")
        cfg.openssl_cafile = data_path("ca.pem")
        cfg.openssl_passphrase = "WRONG PASSWORD"

        with broker.Endpoint(cfg) as ep1, \
             broker.Endpoint(cfg) as ep2:

            port = ep1.listen("127.0.0.1", 0)

            # TODO: This correctly generates an exception in CAF, for which I
            # don't know where to catch it.
            r = ep2.peer("127.0.0.1", port, 0)
            self.assertEqual(r, False)

if __name__ == '__main__':
    unittest.main(verbosity=3)



# ping.py

import sys
import time

import broker

event = int(sys.argv[1])

total_sent_ev1 = 0
total_recv_ev1 = 0
first_t = float(time.time())

last_t = first_t
last_sent_ev1 = 0

def printStats(stats):
    t = stats[0]
    dt = stats[1]
    ev1 = stats[1 + event].value

    global total_recv_ev1
    total_recv_ev1 += ev1

    global last_t, last_sent_ev1
    now = time.time()
    # rate = "sending at {:.2f} ev/s, receiving at {:.2f} ev/s".format(total_sent_ev1 / (now - first_t) , total_recv_ev1 / (now - first_t))
    rate = "sending at {:.2f} ev/s, receiving at {:.2f} ev/s".format((total_sent_ev1 - last_sent_ev1) / (now - last_t), ev1 / dt.total_seconds())
    last_t = now
    last_sent_ev1 = total_sent_ev1

    print("{} dt={} ev{}={} (total {} of {}) {}".format(t, dt, event, ev1, total_recv_ev1, total_sent_ev1, rate))

def sendBatch(p, num):
    event_1s = [broker.zeek.Event("event_{}".format(event), [i, "test"]) for i in range(num)]
    for e in event_1s:
        p.publish(e)

    global total_sent_ev1
    total_sent_ev1 += len(event_1s)

def wait(s, t):
    waited = 0

    while True:
        msgs = s.poll()

        for m in msgs:
            e = broker.zeek.Event(m[1])
            if e.name() == "stats_update":
                printStats(e.args()[0])

        time.sleep(0.01)
        waited += 0.01
        if waited >= t:
            break

ep = broker.Endpoint()
s = ep.make_subscriber("/benchmark/stats")
ss = ep.make_status_subscriber(True);
ep.peer("127.0.0.1", 9999)

# Wait until connection is established.
st = ss.get()

if not (type(st) == broker.Status and st.code() == broker.SC.PeerAdded):
    print("could not connect")
    sys.exit(1)

p = ep.make_publisher("/benchmark/events")

while True:
    sendBatch(p, 5000)
    wait(s, .001)

    if ss.available():
        print(ss.get())
        sys.exit(0)



#! /usr/bin/env python3

import http.server as BaseHTTPServer


class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/plain")
        self.end_headers()

        if "/empty" in self.path:
            self.wfile.write(b"")
        else:
            self.wfile.write(b"It works!")

    def do_POST(self):
        self.send_response(200)
        self.send_header("Content-type", "text/plain")
        self.end_headers()

        if "/empty" in self.path:
            self.wfile.write(b"")
        else:
            self.wfile.write(b"It works!")

    def version_string(self):
        return "1.0"

    def date_time_string(self):
        return "July 22, 2013"


if __name__ == "__main__":
    from optparse import OptionParser
    p = OptionParser()
    p.add_option("-a",
                 "--addr",
                 type="string",
                 default="localhost",
                 help=("listen on given address (numeric IP or host name), "
                       "an empty string (the default) means INADDR_ANY"))
    p.add_option("-p", "--port", type="int", default=32123, help="listen on given TCP port number")
    p.add_option("-m",
                 "--max",
                 type="int",
                 default=-1,
                 help="max number of requests to respond to, -1 means no max")
    options, args = p.parse_args()

    httpd = BaseHTTPServer.HTTPServer((options.addr, options.port), MyRequestHandler)
    if options.max == -1:
        httpd.serve_forever()
    else:
        served_count = 0
        while served_count != options.max:
            httpd.handle_request()
            served_count += 1



"""
Changes section references to be the section number
instead of the title of the section.

Slightly adapted from: https://github.com/jterrace/sphinxtr

Copyright (c) 2012, Jeff Terrace
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
"""

from docutils import nodes
import sphinx.domains.std

class CustomStandardDomain(sphinx.domains.std.StandardDomain):

    def __init__(self, env):
        env.settings['footnote_references'] = 'superscript'
        sphinx.domains.std.StandardDomain.__init__(self, env)

    def resolve_xref(self, env, fromdocname, builder,
                     typ, target, node, contnode):
        res = super(CustomStandardDomain, self).resolve_xref(env, fromdocname, builder,
                                                            typ, target, node, contnode)

        if res is None:
            return res

        if typ == 'ref' and not node['refexplicit']:
            docname, labelid, sectname = self.data['labels'].get(target, ('','',''))
            res['refdocname'] = docname

        return res

def doctree_resolved(app, doctree, docname):
    secnums = app.builder.env.toc_secnumbers
    for node in doctree.traverse(nodes.reference):
        if 'refdocname' in node:
            refdocname = node['refdocname']
            if refdocname in secnums:
                secnum = secnums[refdocname]
                toclist = app.builder.env.tocs[refdocname]
                for child in node.traverse():
                    if isinstance(child, nodes.Text):
                        text = child.astext()
                        anchorname = None
                        for refnode in toclist.traverse(nodes.reference):
                            if refnode.astext() == text:
                                anchorname = refnode['anchorname']
                        if anchorname is None:
                            continue
                        num = '.'.join(map(str, secnum[anchorname]))
                        prefix = 'Section '
                        linktext = prefix + num
                        child.parent.replace(child, nodes.Text(linktext))

def setup(app):
    app.add_domain(CustomStandardDomain, override=True)
    app.connect('doctree-resolved', doctree_resolved)



# Zeekctl test plugin that defines a string to write to zeekctl-config.zeek

import ZeekControl.plugin

class ZeekctlConfigTest(ZeekControl.plugin.Plugin):
    def __init__(self):
        super(ZeekctlConfigTest, self).__init__(apiversion=1)

    def name(self):
        return "zeekctlconfigtest"

    def pluginVersion(self):
        return 1

    def init(self):
        return True

    def zeekctl_config(self):
        return 'redef TestVar = "this is a test";\nredef Test="another test";'




try:
    from . import _broker
except ImportError:
    import _broker

import broker

class Event(_broker.zeek.Event):
    def __init__(self, *args):
        if len(args) == 1 and not isinstance(args[0], str):
            # Parse raw broker message as event.
            _broker.zeek.Event.__init__(self, broker.Data.from_py(args[0]))
        else:
            # (name, arg1, arg2, ...)
            _broker.zeek.Event.__init__(self, args[0], broker.Data.from_py(args[1:]))

    def args(self):
        return [broker.Data.to_py(a) for a in _broker.zeek.Event.args(self)]



# sqlite-connect.py

import broker
import sys
import time

with broker.Endpoint() as ep, \
     ep.make_subscriber('/test') as s, \
     ep.make_status_subscriber(True) as ss:

    ep.peer('127.0.0.1', 9999, 1.0)

    st = ss.get();

    if not (type(st) == broker.Status and st.code() == broker.SC.PeerAdded):
        print('could not connect')
        sys.exit(1)

    c = ep.attach_clone('mystore')

    while True:
        time.sleep(1)
        c.increment('foo', 1)
        print(c.get('foo'))



#!/usr/bin/env python3
import sys
import requests
import time
import pprint
import json

def log(id, last):
    res = requests.get("http://localhost:8082/log/%d/%d" % (id, last)).json()
    if not res['log']:
        sys.stdout.write("waiting...\r")
        sys.stdout.flush()
        return last

    for rec in res['log']:
        print ' '.join(rec), '                '
    return last+len(res['log'])

def wait(id):
    print "Waiting for job %d to finish" % id
    last = 0
    while True:
        res = requests.get("http://localhost:8082/result/%d" % id).json()
        last = log(id, last)
        if res['result'] is not None:
            return json.loads(res['result'])
        time.sleep(.2)


def run(action):
    res = requests.get("http://localhost:8082/%s" % action).json()
    print(wait(res['id']))

def call(action):
    out = requests.get("http://localhost:8082/%s" % action).text
    try:
        return json.loads(out)
    except:
        return out

if __name__ == "__main__":
    if sys.argv[1] == 'bg':
        print call(sys.argv[2])
    else:
        run(sys.argv[1])




import unittest

import broker

class TestTopic(unittest.TestCase):
    def test_string(self):
        t = broker.Topic("/a/b/c")
        self.assertEqual(str(t), "/a/b/c")
        self.assertEqual(t.string(), "/a/b/c")

    def test_append(self):
        t1 = broker.Topic("/a/")
        t2 = broker.Topic("/b/c")
        t3 = t1 / t2
        t2 /= t1

        self.assertEqual(t3.string(), "/a//b/c")
        self.assertEqual(t2.string(), "/b/c/a")

if __name__ == '__main__':
  unittest.main(verbosity=3)



#
# Settings that specify the version of ZeekControl
#

VERSION = "@VERSION@"
ZEEKBASE = "@PREFIX@"
CFGFILE = "@ETC@/zeekctl.cfg"
ZEEKSCRIPTDIR = "@ZEEKSCRIPTDIR@"
LIBDIR = "@LIBDIR@"
LIBDIRINTERNAL = "@PY_MOD_INSTALL_DIR@"



from __future__ import print_function
import unittest
import multiprocessing
import os
import tempfile
import subprocess
import sys
import time

import broker
import broker.zeek

def run_zeek_path():
    base = os.path.realpath(__file__)
    for d in (os.path.join(os.path.dirname(base), "../../build"), os.getcwd()):
        run_zeek = os.path.abspath(os.path.join(d, "tests/python/run-zeek"))
        if os.path.exists(run_zeek):
            return run_zeek

    return "zeek" # Hope for the best ...

ZeekPing = """
redef Broker::default_connect_retry=1secs;
redef Broker::default_listen_retry=1secs;
redef exit_only_after_terminate = T;

global event_count: int = 0;

global ping: event(s: string, c: int);

event zeek_init()
    {
    Broker::subscribe("/test");
    Broker::peer("127.0.0.1", __PORT__/tcp);
    }

function send_event(s: string)
    {
    s += "x";

    if ( event_count == 5 )
        s += "\\x82";

    local e = Broker::make_event(ping, s, event_count);
    Broker::publish("/test", e);
    ++event_count;
    }

event Broker::peer_added(endpoint: Broker::EndpointInfo, s: string)
    {
    send_event("");
    }

event Broker::peer_lost(endpoint: Broker::EndpointInfo, msg: string)
    {
    terminate();
    }

event pong(s: string, n: int)
    {
    send_event(s);
    }
"""

def RunZeek(script, port):
    try:
        tmp = tempfile.NamedTemporaryFile(mode="w", suffix=".zeek", delete=False)
        print(script.replace("__PORT__", str(port)), file=tmp)
        tmp.close()
        subprocess.check_call([run_zeek_path(), "-b", "-B", "broker", tmp.name])
        return True
    except subprocess.CalledProcessError:
        return False
    finally:
        os.unlink(tmp.name)

class TestCommunication(unittest.TestCase):
    def test_ping(self):
        with broker.Endpoint() as ep, \
             ep.make_subscriber("/test") as sub:

            port = ep.listen("127.0.0.1", 0)

            p = multiprocessing.Process(target=RunZeek, args=(ZeekPing, port))
            p.daemon = True
            p.start()

            for i in range(0, 6):
                (t, msg) = sub.get()
                ev = broker.zeek.Event(msg)
                (s, c) = ev.args()
                expected_arg = "x" + "Xx" * i

                if i == 5:
                    expected_arg = expected_arg.encode('utf-8') + b'\x82'

                self.assertEqual(ev.name(), "ping")
                self.assertEqual(s, expected_arg)
                self.assertEqual(c, i)

                if i < 3:
                    ev = broker.zeek.Event("pong", s + "X", c)
                elif i < 5:
                    ev = broker.zeek.Event("pong", s.encode('utf-8') + b'X', c)
                else:
                    ev = broker.zeek.Event("pong", 'done', c)

                ep.publish("/test", ev)

if __name__ == '__main__':
    unittest.main(verbosity=3)



# Zeekctl test plugin that defines custom commands using the old BroControl
# legacy API.

import BroControl.plugin
import BroControl.cmdresult

class CommandTest(BroControl.plugin.Plugin):
    def __init__(self):
        super(CommandTest, self).__init__(apiversion=1)

    def name(self):
        return "commandtest"

    def pluginVersion(self):
        return 1

    def init(self):
        return True

    def commands(self):
        return [("testcmd", "[<nodes>]", "Test command that expects arguments"),
                ("", "", "Another test command")]

    def cmd_custom(self, cmd, args, cmdout):
        results = BroControl.cmdresult.CmdResult()

        # This is an easy way to force the plugin command to return failure.
        if args == "fail":
            results.ok = False
        else:
            results.ok = True

        cmdout.info("Command name: %s" % cmd)
        cmdout.info("Command args: %s" % args)

        return results



# Zeekctl test plugin that defines custom node keys.

import ZeekControl.plugin

class NodeKeysTest(ZeekControl.plugin.Plugin):
    def __init__(self):
        super(NodeKeysTest, self).__init__(apiversion=1)

    def name(self):
        return "nodekeystest"

    def pluginVersion(self):
        return 1

    def init(self):
        with open("keys.out", "w") as f:
            for n in self.nodes():
                f.write("key1: %s\n" % n.nodekeystest_key1)
                f.write("key2: %s\n" % n.nodekeystest_key2)

        return True

    def nodeKeys(self):
        return ["key1", "key2"]



#! /usr/bin/env python3

import sys
import os

cflags = os.environ.get("CFLAGS", "")
os.environ["CFLAGS"] = cflags + " -fno-strict-aliasing"

from setuptools import setup
from distutils.core import Extension

description = """
The PySubnetTree package provides a Python data structure SubnetTree
that maps subnets given in CIDR notation (incl. corresponding IPv6
versions) to Python objects. Lookups are performed by longest-prefix
matching.
"""

with open('README') as file:
    long_description = file.read()

setup(name="pysubnettree",
    version="0.35-4", # Filled in automatically.
    maintainer="The Zeek Project",
    maintainer_email="info@zeek.org",
    license="BSD",
    description=description,
    py_modules=['SubnetTree'],
    url="https://github.com/zeek/pysubnettree",
    ext_modules = [
        Extension("_SubnetTree",
            sources=["SubnetTree.cc", "patricia.c", "SubnetTree_wrap.cc"],
            depends=["include/SubnetTree.h", "include/patricia.h"],
            include_dirs=["include/"])
        ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: BSD :: FreeBSD',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Networking',
        'Topic :: Utilities',
    ],
)




# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.7
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info
if version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_SubnetTree', [dirname(__file__)])
        except ImportError:
            import _SubnetTree
            return _SubnetTree
        if fp is not None:
            try:
                _mod = imp.load_module('_SubnetTree', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _SubnetTree = swig_import_helper()
    del swig_import_helper
else:
    import _SubnetTree
del version_info
try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr_nondynamic(self, class_type, name, static=1):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    if (not static):
        return object.__getattr__(self, name)
    else:
        raise AttributeError(name)

def _swig_getattr(self, class_type, name):
    return _swig_getattr_nondynamic(self, class_type, name, 0)


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object:
        pass
    _newclass = 0


class inx_addr(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, inx_addr, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, inx_addr, name)
    __repr__ = _swig_repr
    __swig_setmethods__["sin"] = _SubnetTree.inx_addr_sin_set
    __swig_getmethods__["sin"] = _SubnetTree.inx_addr_sin_get
    if _newclass:
        sin = _swig_property(_SubnetTree.inx_addr_sin_get, _SubnetTree.inx_addr_sin_set)
    __swig_setmethods__["sin6"] = _SubnetTree.inx_addr_sin6_set
    __swig_getmethods__["sin6"] = _SubnetTree.inx_addr_sin6_get
    if _newclass:
        sin6 = _swig_property(_SubnetTree.inx_addr_sin6_get, _SubnetTree.inx_addr_sin6_set)

    def __init__(self):
        this = _SubnetTree.new_inx_addr()
        try:
            self.this.append(this)
        except:
            self.this = this
    __swig_destroy__ = _SubnetTree.delete_inx_addr
    __del__ = lambda self: None
inx_addr_swigregister = _SubnetTree.inx_addr_swigregister
inx_addr_swigregister(inx_addr)

class SubnetTree(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, SubnetTree, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, SubnetTree, name)
    __repr__ = _swig_repr

    def __init__(self, binary_lookup_mode=False):
        this = _SubnetTree.new_SubnetTree(binary_lookup_mode)
        try:
            self.this.append(this)
        except:
            self.this = this
    __swig_destroy__ = _SubnetTree.delete_SubnetTree
    __del__ = lambda self: None

    def insert(self, *args):
        return _SubnetTree.SubnetTree_insert(self, *args)

    def remove(self, *args):
        return _SubnetTree.SubnetTree_remove(self, *args)

    def lookup(self, *args):
        return _SubnetTree.SubnetTree_lookup(self, *args)

    def prefixes(self, ipv4_native=False, with_len=True):
        return _SubnetTree.SubnetTree_prefixes(self, ipv4_native, with_len)

    def get_binary_lookup_mode(self):
        return _SubnetTree.SubnetTree_get_binary_lookup_mode(self)

    def set_binary_lookup_mode(self, binary_lookup_mode=True):
        return _SubnetTree.SubnetTree_set_binary_lookup_mode(self, binary_lookup_mode)

    def __contains__(self, *args):
        return _SubnetTree.SubnetTree___contains__(self, *args)

    def __getitem__(self, cidr):
        return _SubnetTree.SubnetTree___getitem__(self, cidr)

    def __setitem__(self, cidr, data):
        return _SubnetTree.SubnetTree___setitem__(self, cidr, data)

    def __delitem__(self, cidr):
        return _SubnetTree.SubnetTree___delitem__(self, cidr)
SubnetTree_swigregister = _SubnetTree.SubnetTree_swigregister
SubnetTree_swigregister(SubnetTree)

# This file is compatible with both classic and new-style classes.





#!/usr/bin/env python3

import sys

if len(sys.argv) != 2:
    print("Expected one argument containing the file to clean")
    sys.exit(-1)

with open(sys.argv[1], 'r') as f:

    files = {}
    cur_file = ''
    lines = f.readlines()

    for line in lines:

        if line == 'end_of_record':
            cur_file = ''
            continue

        parts = line.split(':', 1)
        if parts[0] == 'SF':
            cur_file = parts[1].strip()
            while cur_file.find('src/zeek/') != -1:
                cur_file = cur_file.replace('src/zeek/', 'src/', 1)

            if cur_file not in files:
                files[cur_file] = {}
        elif parts[0] == 'DA':
            da_parts = parts[1].split(',')
            line = int(da_parts[0])
            count = int(da_parts[1])

            if files[cur_file].get(line, 0) == 0:
                files[cur_file][line] = count

    for name in files:

        print('TN:')
        print('SF:{}'.format(name))

        das = list(files[name].keys())
        das.sort()

        for da in das:
            print('DA:{},{}'.format(da, files[name][da]))
        print('end_of_record')



"""
    The Zeek domain for Sphinx.
"""

def setup(Sphinx):
    Sphinx.add_domain(ZeekDomain)
    Sphinx.add_node(see)
    Sphinx.add_directive_to_domain('zeek', 'see', SeeDirective)
    Sphinx.connect('doctree-resolved', process_see_nodes)

from sphinx import addnodes
from sphinx.domains import Domain, ObjType, Index
from sphinx.locale import _
from sphinx.directives import ObjectDescription
from sphinx.roles import XRefRole
from sphinx.util.nodes import make_refnode
from sphinx import version_info

from sphinx.util import logging
logger = logging.getLogger(__name__)

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives
from docutils.parsers.rst.roles import set_classes

class see(nodes.General, nodes.Element):
    refs = []

class SeeDirective(Directive):
    has_content = True

    def run(self):
        n = see('')
        n.refs = " ".join(self.content).split()
        return [n]

# Wrapper for creating a tuple for index nodes, staying backwards
# compatible to Sphinx < 1.4:
def make_index_tuple(indextype, indexentry, targetname, targetname2):
    if version_info >= (1, 4, 0, '', 0):
        return (indextype, indexentry, targetname, targetname2, None)
    else:
        return (indextype, indexentry, targetname, targetname2)

def process_see_nodes(app, doctree, fromdocname):
    for node in doctree.traverse(see):
        content = []
        para = nodes.paragraph()
        para += nodes.Text("See also:", "See also:")
        for name in node.refs:
            join_str = " "
            if name != node.refs[0]:
                join_str  = ", "
            link_txt = join_str + name;

            if name not in app.env.domaindata['zeek']['idtypes']:
                # Just create the text and issue warning
                logger.warning('%s: unknown target for ".. zeek:see:: %s"', fromdocname, name, location=node)
                para += nodes.Text(link_txt, link_txt)
            else:
                # Create a reference
                typ = app.env.domaindata['zeek']['idtypes'][name]
                todocname = app.env.domaindata['zeek']['objects'][(typ, name)]

                newnode = nodes.reference('', '')
                innernode = nodes.literal(_(name), _(name), classes=['xref'])
                newnode['refdocname'] = todocname
                newnode['refuri'] = app.builder.get_relative_uri(
                    fromdocname, todocname)
                newnode['refuri'] += '#' + typ + '-' + name
                newnode.append(innernode)
                para += nodes.Text(join_str, join_str)
                para += newnode

        content.append(para)
        node.replace_self(content)

class ZeekGeneric(ObjectDescription):
    option_spec = {
        'source-code': directives.unchanged
    }

    def __init__(self, *args, **kwargs):
        super(ObjectDescription, self).__init__(*args, **kwargs)
        options = args[2]
        self.code_url = None

        if 'source-code' in options and 'zeek-code-url' in self.env.config:
            base_url = self.env.config['zeek-code-url']
            path, start, end = options['source-code'].split()
            path_parts = path.split('/')
            file_name = path_parts[-1]

            # Don't have anything to link to for BIFs
            if not file_name.endswith('.bif.zeek'):
                self.code_url = f'{base_url}/scripts/{path}#L{start}-L{end}'

    def get_obj_name(self):
        return self.objtype

    def update_type_map(self, idname):
        if 'idtypes' not in self.env.domaindata['zeek']:
            self.env.domaindata['zeek']['idtypes'] = {}
        self.env.domaindata['zeek']['idtypes'][idname] = self.get_obj_name()

    def process_signode(self, name, sig, signode, targetname):
        signode['names'].append(targetname)
        signode['ids'].append(targetname)
        signode['first'] = (not self.names)
        self.state.document.note_explicit_target(signode)

    def add_target_and_index(self, name, sig, signode):
        targetname = self.get_obj_name() + '-' + name

        if targetname not in self.state.document.ids:
            self.process_signode(name, sig, signode, targetname)

            objects = self.env.domaindata['zeek']['objects']
            key = (self.get_obj_name(), name)

            if ( key in objects and self.get_obj_name() != "id" and
                 self.get_obj_name() != "type" ):
                logger.warning('%s: duplicate description of %s %s, ' %
                        (self.env.docname, self.get_obj_name(), name) +
                        'other instance in ' +
                        self.env.doc2path(objects[key]),
                        self.lineno)

            objects[key] = self.env.docname
            self.update_type_map(name)

        indextext = self.get_index_text(name)

        if indextext:
            self.indexnode['entries'].append(make_index_tuple('single',
                                             indextext, targetname,
                                             targetname))

    def get_index_text(self, name):
        return _('%s (%s)') % (name, self.get_obj_name())

    def handle_signature(self, sig, signode):
        if self.code_url:
            signode += nodes.reference(sig, sig,
                                       refuri=self.code_url,
                                       reftitle='View Source Code')

            # Could embed snippets directly, but would probably want to clean
            # up how it's done: don't use an external script, figure out why
            # tab/indentation is broken, toggle snippet visibility on mouse
            # hover or other explicit button/link, fix the colors/theming...
            # But for now, leaving this commented out as an example and quick
            # way of checking that the code ranges that Zeekygen outputs are
            # sensible.

            # import urllib
            # snippet_target = urllib.parse.quote(self.code_url, '')
            # snippet_url = 'https://emgithub.com/embed.js'
            # snippet_url += f'?target={snippet_target}'
            # snippet_url += '&style=github'
            # snippet_url += '&showLineNumbers=on'
            # snippet_url += '&showBorder=on'
            # snippet_url += '&ts=4'
            # rawnode = nodes.raw('', f'<script src="{snippet_url}"></script>',
            #                     format='html')
            # signode += rawnode

        else:
            signode += addnodes.desc_name("", sig)

        return sig

class ZeekNamespace(ZeekGeneric):
    def add_target_and_index(self, name, sig, signode):
        targetname = self.get_obj_name() + '-' + name

        if targetname not in self.state.document.ids:
            signode['names'].append(targetname)
            signode['ids'].append(targetname)
            signode['first'] = (not self.names)
            self.state.document.note_explicit_target(signode)

            objects = self.env.domaindata['zeek']['objects']
            key = (self.get_obj_name(), name)
            objects[key] = self.env.docname
            self.update_type_map(name)

        indextext = self.get_index_text(name)
        self.indexnode['entries'].append(make_index_tuple('single', indextext,
                                          targetname, targetname))
        self.indexnode['entries'].append(make_index_tuple('single',
                                          "namespaces; %s" % (sig),
                                          targetname, targetname))

    def get_index_text(self, name):
        return _('%s (namespace); %s') % (name, self.env.docname)

    def handle_signature(self, sig, signode):
        signode += addnodes.desc_name("", sig)
        return sig

class ZeekEnum(ZeekGeneric):
    def add_target_and_index(self, name, sig, signode):
        targetname = self.get_obj_name() + '-' + name

        if targetname not in self.state.document.ids:
            self.process_signode(name, sig, signode, targetname)

            objects = self.env.domaindata['zeek']['objects']
            key = (self.get_obj_name(), name)
            objects[key] = self.env.docname
            self.update_type_map(name)

        indextext = self.get_index_text(name)
        #self.indexnode['entries'].append(make_index_tuple('single', indextext,
        #                                  targetname, targetname))
        m = sig.split()

        if len(m) < 2:
            logger.warning("%s: zeek:enum directive missing argument(s)", self.env.docname)
            return

        if m[1] == "Notice::Type":
            if 'notices' not in self.env.domaindata['zeek']:
                self.env.domaindata['zeek']['notices'] = []
            self.env.domaindata['zeek']['notices'].append(
                                (m[0], self.env.docname, targetname))

        self.indexnode['entries'].append(make_index_tuple('single',
                                          "%s (enum values); %s" % (m[1], m[0]),
                                          targetname, targetname))

    def handle_signature(self, sig, signode):
        m = sig.split()
        name = m[0]
        signode += addnodes.desc_name("", name)
        return name

class ZeekIdentifier(ZeekGeneric):
    def get_index_text(self, name):
        return name

class ZeekNative(ZeekGeneric):
    def handle_signature(self, sig, signode):
        # The run() method is overriden to drop signode anyway in favor of
        # simply adding the index and a target nodes and leaving up
        # to the .rst document to explicitly add things that need to
        # be presented in the final rendering (e.g. a section header)
        self.native_name = sig
        return sig

    def process_signode(self, name, sig, signode, targetname):
        pass

    def run(self):
        ns = super().run()
        index_node = ns[0]
        desc_sig_node = ns[1]

        target_id = self.get_obj_name() + '-' + self.native_name
        target_node = nodes.target('', '', ids=[target_id])
        self.state.document.note_explicit_target(target_node)

        # Replace the description node from Sphinx with a simple target node
        return [index_node, target_node]

class ZeekKeyword(ZeekNative):
    def get_index_text(self, name):
        if name and name[0] == '@':
            return _('%s (directive)') % (name)
        else:
            return _('%s (keyword)') % (name)

class ZeekAttribute(ZeekNative):
    def get_index_text(self, name):
        return _('%s (attribute)') % (name)

class ZeekNativeType(ZeekNative):
    def get_obj_name(self):
        # As opposed to using 'native-type', just imitate 'type'.
        return 'type'

class ZeekNotices(Index):
    """
    Index subclass to provide the Zeek notices index.
    """

    name = 'noticeindex'
    localname = _('Zeek Notice Index')
    shortname = _('notices')

    def generate(self, docnames=None):
        content = {}

        if 'notices' not in self.domain.env.domaindata['zeek']:
            return content, False

        for n in self.domain.env.domaindata['zeek']['notices']:
            modname = n[0].split("::")[0]
            entries = content.setdefault(modname, [])
            entries.append([n[0], 0, n[1], n[2], '', '', ''])

        content = sorted(content.items())

        return content, False

class ZeekDomain(Domain):
    """Zeek domain."""
    name = 'zeek'
    label = 'Zeek'

    object_types = {
        'type':             ObjType(_('type'),             'type'),
        'native-type':      ObjType(_('type'),             'type'),
        'namespace':        ObjType(_('namespace'),        'namespace'),
        'id':               ObjType(_('id'),               'id'),
        'keyword':          ObjType(_('keyword'),          'keyword'),
        'enum':             ObjType(_('enum'),             'enum'),
        'attr':             ObjType(_('attr'),             'attr'),
    }

    directives = {
        'type':             ZeekGeneric,
        'native-type':      ZeekNativeType,
        'namespace':        ZeekNamespace,
        'id':               ZeekIdentifier,
        'keyword':          ZeekKeyword,
        'enum':             ZeekEnum,
        'attr':             ZeekAttribute,
    }

    roles = {
        'type':             XRefRole(),
        'namespace':        XRefRole(),
        'id':               XRefRole(),
        'keyword':          XRefRole(),
        'enum':             XRefRole(),
        'attr':             XRefRole(),
        'see':              XRefRole(),
    }

    indices = [
        ZeekNotices,
    ]

    initial_data = {
        'objects': {},  # fullname -> docname, objtype
    }

    def clear_doc(self, docname):
        to_delete = []

        for (typ, name), doc in self.data['objects'].items():
            if doc == docname:
                to_delete.append((typ, name))

        for (typ, name) in to_delete:
            del self.data['objects'][typ, name]

    def resolve_xref(self, env, fromdocname, builder, typ, target, node,
                     contnode):
        objects = self.data['objects']

        if typ == "see":
            if target not in self.data['idtypes']:
                logger.warning('%s: unknown target for ":zeek:see:`%s`"', fromdocname, target)
                return []
            objtype = self.data['idtypes'][target]
            return make_refnode(builder, fromdocname,
                                        objects[objtype, target],
                                        objtype + '-' + target,
                                        contnode, target + ' ' + objtype)
        else:
            objtypes = self.objtypes_for_role(typ)

            for objtype in objtypes:
                if (objtype, target) in objects:
                    return make_refnode(builder, fromdocname,
                                        objects[objtype, target],
                                        objtype + '-' + target,
                                        contnode, target + ' ' + objtype)
                else:
                    logger.warning('%s: unknown target for ":zeek:%s:`%s`"', fromdocname, typ, target)

    def get_objects(self):
        for (typ, name), docname in self.data['objects'].items():
            yield name, name, typ, docname, typ + '-' + name, 1



from __future__ import print_function
from ZeekControl.state import SqliteState

def test_state_basic():
    s = SqliteState(":memory:")

    assert s.get("key") == None
    s.set("key", "value")
    assert s.get("key") == "value"

    s.set("int", 101)
    assert s.get("int") == 101

    s.set("bool", False)
    assert s.get("bool") == False

def test_state_update():
    s = SqliteState(":memory:")

    s.set("key", "value")
    assert s.get("key") == "value"

    s.set("key", "newvalue")
    assert s.get("key") == "newvalue"

def test_state_setdefault():
    s = SqliteState(":memory:")

    s.set("key", "value")
    assert s.get("key") == "value"

    s.setdefault("key", "newvalue")
    assert s.get("key") == "value"

def test_state_items():
    s = SqliteState(":memory:")
    s.set("a", 1)
    s.set("b", "two")

    d = dict(s.items())
    print(d)

    assert d["a"] == 1
    assert d["b"] == "two"



# Zeekctl test plugin that defines plugin options.

import ZeekControl.plugin

class OptionsTest(ZeekControl.plugin.Plugin):
    def __init__(self):
        super(OptionsTest, self).__init__(apiversion=1)

    def name(self):
        return "optionstest"

    def pluginVersion(self):
        return 1

    def init(self):
        with open("options.out", "w") as f:
            f.write("%s\n" % self.getOption("opt1"))
            f.write("%s\n" % self.getOption("opt2"))
            f.write("%s\n" % self.getOption("opt3"))

        return True

    def options(self):
        return [("opt1", "bool", True, "Boolean option"),
                ("opt2", "string", "test str", "String option"),
                ("opt3", "int", 42, "Int option")]



from __future__ import print_function
from collections import defaultdict
from threading import Thread, Lock
from Queue import Queue
import os
import time
import random
import traceback

from ZeekControl import config
from ZeekControl import version
from ZeekControl.zeekctl import ZeekCtl
from ZeekControl import ser as json

from ZeekControl import web

STOP_RUNNING = object()

class TermUI:
    def __init__(self):
        pass

    def info(self, msg):
        print(msg)
    warn = info

    def error(self, msg):
        print("ERROR", msg)

class Logs:
    def __init__(self):
        self.store = defaultdict(list)

    def append(self, id, stream, txt):
        self.store[id].append((stream, txt))

    def get(self, id, since=0):
        msgs = self.store.get(id) or []
        return msgs[since:]

class Common:
    def dump(self, *args):
        return json.dumps(*args)
    def load(self, msg):
        return json.loads(msg)

class ZeekCtrldWorker(Thread, Common):
    def __init__(self, command_queue):
        self.q = Queue()
        Thread.__init__(self)
        self.daemon = True
        self.command_queue = command_queue
        self._id = None

    def send(self, id, cmd, *args):
        self.q.put((id, cmd, args))

    def run(self):
        #FIXME: deepcopy breaks here if i set ui=self
        self.zeekctl = ZeekCtl(ui=TermUI())
        self.zeekctl.ui = self
        self.zeekctl.controller.ui = self
        self.zeekctl.executor.ui = self
        while True:
            if self.iteration():
                return

    def noop(self, *args, **kwargs):
        return True

    def iteration(self):
        id, cmd, args = self.q.get()
        self._id = id

        if cmd is STOP_RUNNING:
            return True

        func = getattr(self.zeekctl, cmd, self.noop)

        def respond(r):
            self.call("result", id, r)

        if not hasattr(func, 'api_exposed'):
            return respond("invalid function")

        try :
            res = func(*args)
        except Exception as e:
            res = traceback.format_exc()
        respond(res)

    def call(self, func, *args):
        self.command_queue.put((None, func, args))

    def info(self, msg):
        return self.call("info", self._id, msg)

    def error(self, msg):
        return self.call("err", self._id, msg)
    warn = error

class ZeekCtld(Thread, Common):
    def __init__(self, command_queue, logs):
        Thread.__init__(self)
        self.daemon = True
        self.logs = logs
        self.command_queue = command_queue
        self.worker = ZeekCtrldWorker(self.command_queue)

        self.results = {}
        self.running = True

        self.id_gen = iter(range(10000000)).next

        self.init()

    def init(self):
        self.worker.start()

    def recv(self):
        msg = self.command_queue.get()
        print("Received", msg)
        return msg

    def run(self):
        return self._run()

    def handle_result(self, id, result):
        print("Got result id=%r result=%r" % (id, result))
        self.results[id] = result
        return "ok"

    def handle_out(self, id, txt):
        print("Got %s id=%r result=%r" % ('out', id, txt))
        self.logs.append(id, 'out', txt)
        return "ok"

    def handle_info(self, id, txt):
        print("Got %s id=%r result=%r" % ('info', id, txt))
        self.logs.append(id, 'info', txt)
        return "ok"

    def handle_err(self, id, txt):
        print("Got %s id=%r result=%r" % ('err', id, txt))
        self.logs.append(id, 'err', txt)
        return "ok"

    def handle_getresult(self, id):
        result = self.results.get(id)
        if result:
            del self.results[id]
        print("sending result=%r for id=%r" % (result, id))
        return result

    def handle_getlog(self, id, since):
        result = self.logs.get(id, since)
        print("sending log=%r for id=%r" % (result, id))
        return result

    def _run(self):
        while self.running:
            (rq, cmd, args) = self.recv()
            func = getattr(self, 'handle_' + cmd, None)
            if func:
                res = func(*args)
                if rq:
                    rq.put(res)
            else:
                t_id = self.send_to_worker(cmd, *args)
                rq.put(t_id)

    def send_to_worker(self, cmd, *args):
        t_id = self.id_gen()
        self.worker.send(t_id, cmd, *args)
        print("started id=%r cmd=%r args=%r" % (t_id, cmd, args))
        return t_id

class Client(Common):
    def __init__(self, command_queue):
        self.command_queue = command_queue

    def call(self, func, *args):
        rq = Queue()
        self.command_queue.put((rq, func, args))
        return rq.get()

    def sync_call(self, func, *args):
        id = self.call(func, *args)
        print("got id", id)
        while True:
            res = self.getresult(id)
            if res:
                return res
            time.sleep(0.1)

    def result(self, id, result):
        return self.call("result", id, result)

    def getresult(self, id):
        return self.call("getresult", id)

    def getlog(self, id, since=0):
        return self.call("getlog", id, since)

def main(basedir='/zeek'):
    logs = Logs()

    command_queue = Queue()
    d = ZeekCtld(command_queue, logs)

    d.start()

    c = Client(command_queue)

    ww = Thread(target=web.run_app, args=[c])
    ww.start()
    print("I Got:", c.sync_call('status'))
    for x in range(20):
        time.sleep(1)

if __name__ == "__main__":
    main()



# Functions to install files on all nodes.

import os
import binascii

from ZeekControl import util
from ZeekControl import config

# In all paths given in this file, ${<option>} will replaced with the value of
# the corresponding configuration option.

# Directories/files in form (path, mirror, optional) which are synced from the manager to
# all remote hosts.
# If "mirror" is False, then the path is assumed to be a directory and it will
# just be created on the remote host.  If "mirror" is True, then the path
# is fully mirrored recursively.  If "optional" is True, then it's ok for
# the path to not exist on the manager.
def get_syncs():
    syncs = [
    ("${zeekbase}", False, False),
    ("${zeekbase}/share", True, False),
    ("${cfgdir}", True, False),
    ("${libdir}", True, True),
    ("${libdir64}", True, True),
    ("${bindir}", True, False),
    ("${policydirsiteinstall}", True, False),
    ("${policydirsiteinstallauto}", True, False),
    # ("${policydir}", True, False),
    # ("${staticdir}", True, False),
    ("${logdir}", False, False),
    ("${spooldir}", False, False),
    ("${tmpdir}", False, False),
    ("${zeekctlconfigdir}/zeekctl-config.sh", True, False)
    ]

    return syncs

# In NFS-mode, only these will be synced.
def get_nfssyncs():
    nfssyncs = [
    ("${spooldir}", False, False),
    ("${tmpdir}", False, False),
    ("${policydirsiteinstall}", True, False),
    ("${policydirsiteinstallauto}", True, False),
    ("${zeekctlconfigdir}/zeekctl-config.sh", True, False)
    ]

    return nfssyncs

# Determine relative cfg_path
def splitall(path):
    """Break out all parts of a path into a list"""
    allparts = []
    while True:
        parts = os.path.split(path)
        if parts[0] == path:
            allparts.insert(0, parts[0])
            break
        if parts[1] == path:
            allparts.insert(0, parts[1])
            break
        path = parts[0]
        allparts.insert(0, parts[1])
    return allparts

def relpath(src, dst):
    """Calculate the relative path to dst)"""
    srcparts = splitall(src)
    dstparts = splitall(dst)
    while srcparts and dstparts and srcparts[0] == dstparts[0]:
        srcparts.pop(0)
        dstparts.pop(0)
    relparts = (len(dstparts) - 1) * ['..'] + srcparts
    return os.path.join(*relparts)

# Generate a shell script "zeekctl-config.sh" that sets env. vars. that
# correspond to zeekctl config options.
def make_zeekctl_config_sh(cmdout):
    ostr = ""
    for (varname, value) in config.Config.options(dynamic=False):
        if isinstance(value, bool):
            # Convert bools to the string "1" or "0"
            value = "1" if value else "0"
        else:
            value = str(value)

        # In order to prevent shell errors, here we convert plugin
        # option names to use underscores, and double quotes in the value
        # are escaped.
        ostr += '%s="%s"\n' % (varname.replace(".", "_"), value.replace('"', '\\"'))

    # Rather than just overwriting the file, we first write out a tmp file,
    # and then rename it to avoid a race condition where a process outside of
    # zeekctl (such as archive-log) is trying to read the file while it is
    # being written.
    cfg_path = os.path.join(config.Config.zeekctlconfigdir, "zeekctl-config.sh")
    tmp_path = os.path.join(config.Config.zeekctlconfigdir, ".zeekctl-config.sh.tmp")

    try:
        with open(tmp_path, "w") as out:
            out.write(ostr)
    except IOError as e:
        cmdout.error("failed to write file: %s" % e)
        return False

    try:
        os.rename(tmp_path, cfg_path)
    except OSError as e:
        cmdout.error("failed to rename file %s: %s" % (tmp_path, e))
        return False

    symlink = os.path.join(config.Config.scriptsdir, "zeekctl-config.sh")

    # Use a relative path instead of absolute (at request/usage of the FreeBSD
    # port, see https://github.com/zeek/zeekctl/pull/22)
    sym_cfg_path = relpath(cfg_path, symlink)

    # check if the symlink needs to be updated
    try:
        update_link = not os.path.islink(symlink) or os.readlink(symlink) != sym_cfg_path
    except OSError as e:
        cmdout.error("failed to read symlink: %s" % e)
        return False

    if update_link:
        # attempt to update the symlink
        try:
            util.force_symlink(sym_cfg_path, symlink)
        except OSError as e:
            cmdout.error("failed to update symlink '%s' to point to '%s': %s" % (symlink, sym_cfg_path, e.strerror))
            return False

    return True


# Create Zeek-side zeekctl configuration file.
def make_layout(path, cmdout, silent=False):
    class Port:
        def __init__(self, startport):
            # This is the first port number to use.
            self.p = startport

        # Record the port number that the specified node will use (if node is
        # None, then don't record it) and return that port number.
        def use_port(self, node):
            port = self.p
            # Increment the port number, since we're using the current one.
            self.p += 1

            if node is not None:
                node.setPort(port)

            return port

    manager = config.Config.manager()
    zeekport = Port(config.Config.zeekport)

    if config.Config.standalone:
        if not silent:
            cmdout.info("generating standalone-layout.zeek ...")

        filename = os.path.join(path, "standalone-layout.zeek")

        ostr = "# Automatically generated. Do not edit.\n"
        # This is the port that standalone nodes listen on for remote
        # control by default.
        ostr += "redef Broker::default_port = %s/tcp;\n" % zeekport.use_port(manager)
        ostr += "event zeek_init()\n"
        ostr += "\t{\n"
        ostr += "\tif ( getenv(\"ZEEKCTL_DISABLE_LISTEN\") == \"\" )\n"
        ostr += "\t\tBroker::listen();\n"
        ostr += "\t}\n"

    else:
        if not silent:
            cmdout.info("generating cluster-layout.zeek ...")

        filename = os.path.join(path, "cluster-layout.zeek")
        workers = config.Config.workers()
        proxies = config.Config.proxies()
        loggers = config.Config.loggers()

        # If no loggers are defined, then manager does the logging.
        manager_is_logger = "F" if loggers else "T"

        ostr = "# Automatically generated. Do not edit.\n"
        ostr += "redef Cluster::manager_is_logger = %s;\n" % manager_is_logger
        ostr += "redef Cluster::nodes = {\n"

        # Control definition.  For now just reuse the manager information.
        ostr += '\t["control"] = [$node_type=Cluster::CONTROL, $ip=%s, $p=%s/tcp],\n' % (util.format_zeek_addr(manager.addr), zeekport.use_port(None))

        # Loggers definition
        for lognode in loggers:
            ostr += '\t["%s"] = [$node_type=Cluster::LOGGER, $ip=%s, $p=%s/tcp],\n' % (lognode.name, util.format_zeek_addr(lognode.addr), zeekport.use_port(lognode))

        # Manager definition
        ostr += '\t["%s"] = [$node_type=Cluster::MANAGER, $ip=%s, $p=%s/tcp],\n' % (manager.name, util.format_zeek_addr(manager.addr), zeekport.use_port(manager))

        # Proxies definition (all proxies use same logger as the manager)
        for p in proxies:
            ostr += '\t["%s"] = [$node_type=Cluster::PROXY, $ip=%s, $p=%s/tcp, $manager="%s"],\n' % (p.name, util.format_zeek_addr(p.addr), zeekport.use_port(p), manager.name)

        # Workers definition
        for w in workers:
            p = w.count % len(proxies)
            ostr += '\t["%s"] = [$node_type=Cluster::WORKER, $ip=%s, $p=%s/tcp, $interface="%s", $manager="%s"],\n' % (w.name, util.format_zeek_addr(w.addr), zeekport.use_port(w), w.interface, manager.name)

        # Activate time-machine support if configured.
        if config.Config.timemachinehost:
            ostr += '\t["time-machine"] = [$node_type=Cluster::TIME_MACHINE, $ip=%s, $p=%s],\n' % (config.Config.timemachinehost, config.Config.timemachineport)

        ostr += "};\n"

    try:
        with open(filename, "w") as out:
            out.write(ostr)
    except IOError as e:
        cmdout.error("failed to write file: %s" % e)
        return False

    return True


# Reads in a list of networks from file.
def read_networks(fname):

    nets = []

    with open(fname, "r") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            fields = line.split(None, 1)

            cidr = util.format_zeek_prefix(fields[0])
            tag = fields[1] if len(fields) == 2 else ""

            nets += [(cidr, tag)]

    return nets


# Create Zeek script which contains a list of local networks.
def make_local_networks(path, cmdout):

    netcfg = config.Config.localnetscfg

    try:
        nets = read_networks(netcfg)
    except IndexError:
        cmdout.error("invalid CIDR notation in file: %s" % netcfg)
        return False
    except IOError as e:
        cmdout.error("failed to read file: %s" % e)
        return False

    ostr = "# Automatically generated. Do not edit.\n\n"

    ostr += "redef Site::local_nets = {\n"
    for (cidr, tag) in nets:
        ostr += "\t%s," % cidr
        if tag:
            ostr += "\t# %s" % tag
        ostr += "\n"
    ostr += "};\n\n"

    try:
        with open(os.path.join(path, "local-networks.zeek"), "w") as out:
            out.write(ostr)
    except IOError as e:
        cmdout.error("failed to write file: %s" % e)
        return False

    return True


def make_zeekctl_config_policy(path, cmdout, plugin_reg):
    ostr = '# Automatically generated. Do not edit.\n'
    ostr += 'redef Notice::mail_dest = "%s";\n' % config.Config.mailto
    ostr += 'redef Notice::mail_dest_pretty_printed = "%s";\n' % config.Config.mailalarmsto
    ostr += 'redef Notice::sendmail = "%s";\n' % config.Config.sendmail
    ostr += 'redef Notice::mail_subject_prefix = "%s";\n' % config.Config.mailsubjectprefix
    ostr += 'redef Notice::mail_from = "%s";\n' % config.Config.mailfrom
    ostr += 'redef Broker::table_store_db_directory = "%s";\n' % config.Config.brokerdbdir
    if not config.Config.standalone:
        loggers = config.Config.loggers()
        ntype = "LOGGER" if loggers else "MANAGER"
        ostr += '@if ( Cluster::local_node_type() == Cluster::%s )\n' % ntype

    ostr += 'redef Log::default_rotation_interval = %s secs;\n' % config.Config.logrotationinterval
    ostr += 'redef Log::default_mail_alarms_interval = %s secs;\n' % config.Config.mailalarmsinterval

    if not config.Config.standalone:
        ostr += '@endif\n'

    ostr += 'redef Pcap::snaplen = %s;\n' % config.Config.pcapsnaplen
    ostr += 'redef Pcap::bufsize = %s;\n' % config.Config.pcapbufsize

    seed_str = make_global_hash_seed()
    ostr += 'redef global_hash_seed = "%s";\n' % seed_str

    ostr += 'redef Cluster::default_store_dir = "%s";\n' % config.Config.defaultstoredir

    ostr += plugin_reg.getZeekctlConfig(cmdout)

    if config.Config.compresslogsinflight > 0:
        ostr += 'redef LogAscii::gzip_level = %s;\n' % config.Config.compresslogsinflight
        ostr += 'redef LogAscii::gzip_file_extension = "%s";\n' % config.Config.compressextension

    filename = os.path.join(path, "zeekctl-config.zeek")
    try:
        with open(filename, "w") as out:
            out.write(ostr)
    except IOError as e:
        cmdout.error("failed to write file: %s" % e)
        return False

    return True


# Create a new random seed value if one is not found in the state database (this
# ensures a consistent value after a restart).  Return a string representation.
def make_global_hash_seed():
    seed_str = config.Config.get_state("global-hash-seed")

    if not seed_str:
        # Get 4 bytes of random data (Zeek uses 4 bytes to create an initial
        # seed in the Hasher::MakeSeed() function if the Zeek script constant
        # global_hash_seed is an empty string).
        seed = os.urandom(4)

        # Convert each byte of seed value to a two-digit hex string.
        seed_str = binascii.hexlify(seed)
        seed_str = seed_str.decode()

        config.Config.set_state("global-hash-seed", seed_str)

    return seed_str



#!/usr/bin/env python

# Indents a CAF log with trace verbosity. The script does *not* deal with a log
# with multiple threads.

# usage (read file): indent_trace_log.py FILENAME
#      (read stdin): indent_trace_log.py -

import argparse, sys, os, re

def is_entry(line):
    return 'TRACE' in line and 'ENTRY' in line

def is_exit(line):
    return 'TRACE' in line and 'EXIT' in line

def print_indented(line, indent):
    if is_exit(line):
        indent = indent[:-2]
    sys.stdout.write(indent)
    sys.stdout.write(line)
    if is_entry(line):
        indent += "  "
    return indent

def read_lines(fp, ids):
    indent = ""
    if not ids or len(ids) == 0:
        for line in fp:
            indent = print_indented(line, indent)
    else:
        rx = re.compile('.+ (?:actor|ID = )([0-9]+) .+')
        for line in fp:
            rx_res = rx.match(line)
            if rx_res != None and rx_res.group(1) in ids:
                indent = print_indented(line, indent)

def main():
    parser = argparse.ArgumentParser(description='Add a new C++ class.')
    parser.add_argument('-i', dest='ids', action='append', help='only include actors with given ID(s)')
    parser.add_argument("log", help='path to the log file or "-" for reading from STDIN')
    args = parser.parse_args()
    filepath = args.log
    if filepath == '-':
        read_lines(sys.stdin, args.ids)
    else:
        if not os.path.isfile(filepath):
            sys.exit()
        with open(filepath) as fp:
            read_lines(fp, args.ids)

if __name__ == "__main__":
    main()



#
# Testing the data conversion back and forth between broker.Data and Python.
# Note that the user will rarely see broker.Data instances directly. We make
# sure that the transparent conversion works as expected.
#

import datetime
import time
import math
import ipaddress
import sys
import types
import unittest

import broker


class TestDataConstruction(unittest.TestCase):

    # Given a Python value 'p' and expected broker type 't', convert 'p' to
    # the corresponding broker type and verify that it has the correct type
    # and the expected string representation 's' (note this is not necessarily
    # the same as str(p)).
    def check_to_broker(self, p, s, t):
        assert not isinstance(p, broker.Data)
        b = broker.Data(p)
        #print("[to_broker] ({} / {}) -> ({} / {}) (expected: {} / {})".format(p, type(p), str(b), b.get_type(), s, t))

        if s is not None:
            self.assertEqual(str(b), s)

        self.assertEqual(b.get_type(), t)
        return b

    # Given a broker value 'b' and a Python value 'p', convert 'b' to the
    # corresponding python type, then verify that the type and value match
    # the original python value 'p'.
    def check_to_py(self, b, p):
        b2p = broker.Data.to_py(b)
        #print("[to_py] data({} / {}) -> ({} / {})".format(str(b), b.get_type(), b2p, type(b2p)))

        self.assertIsInstance(b2p, type(p))

        if isinstance(p, datetime.datetime):
            # Some pythons may have rounding issues converting datetime to
            # timestamp and back again.  https://bugs.python.org/issue23517

            if p.tzinfo:
                # Our test cases were explicit in using UTC
                self.assertEqual(b2p.tzinfo, p.tzinfo)

                self.assertEqual(b2p.year, p.year)
                self.assertEqual(b2p.month, p.month)
                self.assertEqual(b2p.day, p.day)
                self.assertEqual(b2p.hour, p.hour)
                self.assertEqual(b2p.minute, p.minute)
                self.assertEqual(b2p.second, p.second)

                us_equal = (b2p.microsecond == p.microsecond or
                            b2p.microsecond == p.microsecond - 1 or
                            b2p.microsecond == p.microsecond + 1)
                self.assertTrue(us_equal)
            else:
                # 'b2p' is in UTC and 'p' is assumed to be local time
                self.assertEqual(b2p.tzinfo, broker.utc)

                b2p_us = b2p.microsecond

                b2p_ts  = (b2p - datetime.datetime(1970, 1, 1, tzinfo=broker.utc)).total_seconds()
                b2p_ts = math.trunc(b2p_ts)

                p_us = p.microsecond

                p_ts = time.mktime(p.timetuple())

                self.assertEqual(b2p_ts, p_ts)

                us_equal = (b2p_us == p_us or
                            b2p_us == p_us - 1 or
                            b2p_us == p_us + 1)
                self.assertTrue(us_equal)

        else:
            self.assertEqual(b2p, p)

    def check_to_broker_and_back(self, p, s, t, p_final=None):
        """Given a Python value p, convert to Broker and check that the latter renders
        to string s and matches Broker type t. Then convert back to Python and
        verify that type and value match p_final. If p_final is None, use the
        original input p.
        """
        b = self.check_to_broker(p, s, t)
        self.check_to_py(b, p if p_final is None else p_final)
        return b

    def test_bool(self):
        self.check_to_broker_and_back(True, 'T', broker.Data.Type.Boolean)
        self.check_to_broker_and_back(False, 'F', broker.Data.Type.Boolean)

    def test_integer(self):
        self.check_to_broker_and_back(42, '42', broker.Data.Type.Integer)

        self.check_to_broker_and_back(-42, '-42', broker.Data.Type.Integer)

        # Test a value that is beyond range of unsigned 32-bit integer
        self.check_to_broker_and_back(5123123123, '5123123123', broker.Data.Type.Integer)

    def test_count(self):
        self.check_to_broker_and_back(broker.Count(42), '42', broker.Data.Type.Count)

        # Test a value that is beyond range of unsigned 32-bit integer
        self.check_to_broker_and_back(broker.Count(5123123123), '5123123123', broker.Data.Type.Count)

    def test_count_overflow(self):
        with self.assertRaises(Exception) as context:
            # I've seen this raise either OverflowError or SystemError
            # depending on Python version is seems.
            self.check_to_broker(broker.Count(-1), '-1', broker.Data.Type.Count)

    def test_real(self):
        self.check_to_broker_and_back(1e18, '1000000000000000000.000000', broker.Data.Type.Real)
        self.check_to_broker_and_back(4.2, '4.200000', broker.Data.Type.Real)
        self.check_to_broker_and_back(-4.2, '-4.200000', broker.Data.Type.Real)

    def test_timespan(self):
        to_us = lambda x: x.microseconds + (x.seconds + x.days * 24 * 3600) * 10**6
        to_ns = lambda x: to_us(x) * 10**3

        # Setup timespan values
        neg1ms = datetime.timedelta(microseconds = -1000)
        neg1ms_ns = -1000 * 10**3
        self.assertEqual(neg1ms_ns, to_ns(neg1ms))

        neg42sec = datetime.timedelta(milliseconds = -42 * 10**3)
        neg42sec_ns = (-42 * 10**6) * 10**3
        self.assertEqual(neg42sec_ns, to_ns(neg42sec))

        pos1ms2us = datetime.timedelta(milliseconds = 1, microseconds = 2)
        pos1ms2us_ns = (1000 + 2) * 10**3
        self.assertEqual(pos1ms2us_ns, to_ns(pos1ms2us))

        pos1day2s3us = datetime.timedelta(days = 1, seconds = 2, microseconds = 3)
        pos1day2s3us_ns = (3 + (2 + 24 * 3600) * 10**6) * 10**3
        self.assertEqual(pos1day2s3us_ns, to_ns(pos1day2s3us))

        # Verify Timespan only
        self.assertEqual(broker.Timespan(neg1ms_ns), broker.Timespan(to_ns(neg1ms)))
        self.assertEqual(broker.Timespan(neg42sec_ns), broker.Timespan(to_ns(neg42sec)))
        self.assertEqual(broker.Timespan(pos1ms2us_ns), broker.Timespan(to_ns(pos1ms2us)))
        self.assertEqual(broker.Timespan(pos1day2s3us_ns), broker.Timespan(to_ns(pos1day2s3us)))

        # Verify Data
        self.check_to_broker_and_back(neg1ms, '-1000000ns', broker.Data.Type.Timespan)
        self.check_to_broker_and_back(neg42sec, '-42000000000ns', broker.Data.Type.Timespan)
        self.check_to_broker_and_back(pos1ms2us, '1002000ns', broker.Data.Type.Timespan)
        self.check_to_broker_and_back(pos1day2s3us, '86402000003000ns', broker.Data.Type.Timespan)

    def test_timestamp(self):
        self.check_to_broker(broker.now(), None, broker.Data.Type.Timestamp)

        today = datetime.datetime.now(broker.utc)
        self.check_to_broker_and_back(today, None, broker.Data.Type.Timestamp)

        today = datetime.datetime.now(datetime.timezone.utc)
        self.check_to_broker_and_back(today, None, broker.Data.Type.Timestamp)

        today = datetime.datetime.now()
        self.check_to_broker_and_back(today, None, broker.Data.Type.Timestamp)

        past = datetime.datetime(1970, 1, 31, tzinfo=broker.utc)
        self.check_to_broker_and_back(past, None, broker.Data.Type.Timestamp)

        # Test a time value with number of seconds since Jan. 1 1970 beyond
        # the range of a signed 32-bit integer (the "year 2038 problem").
        future = datetime.datetime(2040, 1, 31, tzinfo=broker.utc)
        try:
            self.check_to_broker_and_back(future, None, broker.Data.Type.Timestamp)
        except OverflowError:
            # This test fails on some 32-bit systems (such as Debian 9 i386),
            # but for now we just ignore this failure.
            pass

    def test_string(self):
        self.check_to_broker_and_back('', '', broker.Data.Type.String)
        self.check_to_broker_and_back('foo', 'foo', broker.Data.Type.String)
        self.check_to_broker_and_back('\ttab', '\ttab', broker.Data.Type.String)
        self.check_to_broker_and_back('new\n', 'new\n', broker.Data.Type.String)

    def test_address_v4(self):
        addr = ipaddress.IPv4Address('0.0.0.0')
        self.check_to_broker_and_back(addr, '0.0.0.0', broker.Data.Type.Address)

        addr = ipaddress.IPv4Address('1.2.3.4')
        self.check_to_broker_and_back(addr, '1.2.3.4', broker.Data.Type.Address)

        addr = ipaddress.IPv4Address('255.255.255.255')
        self.check_to_broker_and_back(addr, '255.255.255.255', broker.Data.Type.Address)

    def test_address_v6(self):
        addr = ipaddress.IPv6Address('::')
        self.check_to_broker_and_back(addr, '::', broker.Data.Type.Address)

        addr = ipaddress.IPv6Address('::1')
        self.check_to_broker_and_back(addr, '::1', broker.Data.Type.Address)

        addr = ipaddress.IPv6Address('1:2:3:4:5:6:7:8')
        self.check_to_broker_and_back(addr, '1:2:3:4:5:6:7:8', broker.Data.Type.Address)

        addr = ipaddress.IPv6Address('ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff')
        self.check_to_broker_and_back(addr, 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff', broker.Data.Type.Address)

    def test_subnet_v4(self):
        sn = ipaddress.IPv4Network('1.2.3.4/32')
        self.check_to_broker_and_back(sn, '1.2.3.4/32', broker.Data.Type.Subnet)

        sn = ipaddress.IPv4Network('10.0.0.0/8')
        self.check_to_broker_and_back(sn, '10.0.0.0/8', broker.Data.Type.Subnet)

        sn = ipaddress.IPv4Network('0.0.0.0/0')
        self.check_to_broker_and_back(sn, '0.0.0.0/0', broker.Data.Type.Subnet)

    def test_subnet_v6(self):
        sn = ipaddress.IPv6Network('::1/128')
        self.check_to_broker_and_back(sn, '::1/128', broker.Data.Type.Subnet)

        sn = ipaddress.IPv6Network('fc00::/7')
        self.check_to_broker_and_back(sn, 'fc00::/7', broker.Data.Type.Subnet)

        sn = ipaddress.IPv6Network('::/0')
        self.check_to_broker_and_back(sn, '::/0', broker.Data.Type.Subnet)

    def test_port(self):
        self.check_to_broker_and_back(broker.Port(65535, broker.Port.TCP), "65535/tcp", broker.Data.Type.Port)
        self.check_to_broker_and_back(broker.Port(53, broker.Port.UDP), "53/udp", broker.Data.Type.Port)
        self.check_to_broker_and_back(broker.Port(8, broker.Port.ICMP), "8/icmp", broker.Data.Type.Port)
        self.check_to_broker_and_back(broker.Port(0, broker.Port.Unknown), "0/?", broker.Data.Type.Port)

    def _test_set_impl(self, set_itype, set_otype=None):
        # Common set testing functionality for an input type into Broker and a
        # corresponding output type (set_itype/set_otype). When the output type
        # isn't provided, use the input type:
        set_otype = set_itype if set_otype is None else set_otype

        # Test an empty set
        self.check_to_broker_and_back(set_itype(), '{}', broker.Data.Type.Set, set_otype())

        # Test a simple set
        pi, po = set_itype([1, 2, 3]), set_otype([1, 2, 3])
        d = self.check_to_broker_and_back(pi, '{1, 2, 3}', broker.Data.Type.Set, po)

        for (i, x) in enumerate(d.as_set()):
            self.check_to_broker(x, str(i + 1), broker.Data.Type.Integer)
            self.check_to_py(x, i + 1)

        # Test a set that contains various data types
        d = broker.Data(set_itype(['foo', ipaddress.IPv6Address('::1'), None]))
        for (i, x) in enumerate(d.as_set()):
            if i == 1:
                self.check_to_broker(x, 'foo', broker.Data.Type.String)
                self.check_to_py(x, 'foo')
            elif i == 2:
                self.check_to_broker(x, '::1', broker.Data.Type.Address)
                self.check_to_py(x, ipaddress.IPv6Address('::1'))

        # Test some of our own methods on wrapped sets.
        d = broker.Data(set_itype([1, 2, 3])).as_set()
        self.assertEqual(str(d), "Set{1, 2, 3}")
        d.remove(broker.Data(2))
        self.assertEqual(str(d), "Set{1, 3}")
        d.clear()
        self.assertEqual(str(d), "Set{}")

    def test_set(self):
        self._test_set_impl(set)

    def test_frozenset(self):
        # Python frozensets convert to Broker sets, so by default their mapping
        # back to Python results in regular sets.
        self._test_set_impl(frozenset, set)

    def _test_table_impl(self, table_itype, table_otype=None):
        # Common table testing functionality for an input type into Broker and a
        # corresponding output type (table_itype/table_otype). When the output
        # type isn't provided, use the input type:
        table_otype = table_itype if table_otype is None else table_otype

        # Test an empty table
        self.check_to_broker_and_back(table_itype({}), '{}', broker.Data.Type.Table, table_otype())

        # Test a simple table
        d = {"a": 1, "b": 2, "c": 3}
        pi, po = table_itype(d), table_otype(d)
        d = self.check_to_broker_and_back(pi, '{a -> 1, b -> 2, c -> 3}', broker.Data.Type.Table, po)

        for (i, (k, v)) in enumerate(d.as_table().items()):
            self.check_to_broker(k, ["a", "b", "c"][i], broker.Data.Type.String)
            self.check_to_py(k, ["a", "b", "c"][i])
            self.check_to_broker(v, str(i + 1), broker.Data.Type.Integer)
            self.check_to_py(v, i + 1)

        # Test a table that contains different data types
        p = table_itype({
            True: 42,
            broker.Port(22, broker.Port.TCP): False,
            (1,2,3): [4,5,6],
            broker.Count(13): "test",
        })
        d = self.check_to_broker(
            p, '{T -> 42, 13 -> test, 22/tcp -> F, (1, 2, 3) -> (4, 5, 6)}',
            broker.Data.Type.Table)

        t = d.as_table()

        self.check_to_broker(t[broker.Data(True)], "42", broker.Data.Type.Integer)
        self.check_to_py(t[broker.Data(True)], 42)

        self.check_to_broker(t[broker.Data(broker.Port(22, broker.Port.TCP))], "F", broker.Data.Type.Boolean)
        self.check_to_py(t[broker.Data(broker.Port(22, broker.Port.TCP))], False)

        self.check_to_broker(t[broker.Data((1, 2, 3))], "(4, 5, 6)", broker.Data.Type.Vector)
        self.check_to_py(t[broker.Data([1, 2, 3])], (4, 5, 6))

        self.check_to_broker(t[broker.Data(broker.Count(13))], "test", broker.Data.Type.String)
        self.check_to_py(t[broker.Data(broker.Count(13))], "test")

    def test_dict(self):
        self._test_table_impl(dict)

    def test_mapping_proxy_type(self):
        # Python MappingProxyType instances will convert to Broker tables, which
        # by default convert back to regular dicts.
        self._test_table_impl(types.MappingProxyType, dict)

    def test_vector(self):
        # Test an empty vector
        self.check_to_broker_and_back((), '()', broker.Data.Type.Vector)

        # Test a simple vector
        d = self.check_to_broker((1, 2, 3), '(1, 2, 3)', broker.Data.Type.Vector)
        # Either a list or a tuple are mapped to a Broker vector.
        d = self.check_to_broker([1, 2, 3], '(1, 2, 3)', broker.Data.Type.Vector)

        for (i, x) in enumerate(d.as_vector()):
            self.check_to_broker(x, str(i + 1), broker.Data.Type.Integer)
            self.check_to_py(x, i + 1)

        # Test a vector that contains various data types
        d = broker.Data(['foo', [[True]], ipaddress.IPv6Address('::1'), None])
        v = d.as_vector()
        self.check_to_broker(v[0], 'foo', broker.Data.Type.String)
        self.check_to_py(v[0], 'foo')

        v1 = v[1].as_vector()
        self.check_to_broker(v1, '((T))', broker.Data.Type.Vector)
        self.check_to_py(v[1], ((True,),))
        self.check_to_broker(v1[0], '(T)', broker.Data.Type.Vector)
        self.check_to_py(v1[0], (True,))

        self.check_to_broker(v[2], '::1', broker.Data.Type.Address)
        self.check_to_py(v[2], ipaddress.IPv6Address("::1"))

        self.check_to_broker(v[3], 'nil', broker.Data.Type.Nil)

if __name__ == '__main__':
    unittest.main(verbosity=3)
